import ast
import json
import logging
import threading
import time

from http import HTTPStatus
from itertools import zip_longest
from pathlib import Path
from rauth import OAuth2Service
from requests import ConnectionError
from requests.adapters import HTTPAdapter
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from urllib3.util.retry import Retry
from urllib.parse import urljoin, urlparse
from urllib.error import URLError

import iridauploader.model as model
# For a truly independent api module, we should have a signal, or pubsub system in the module, that the progress module
# can subscribe to. That way, the api module is seperate, and other applications could use the emits/messages in their
# own setups.
import iridauploader.progress as progress

from . import exceptions

# These strings are used to determine which upload mode is being used when uploading sequence files
# They are included in the `api` __init__.py s.t. they can be used by the other modules without interacting with the
# api layer
MODE_DEFAULT = "default"
MODE_ASSEMBLIES = "assemblies"
MODE_FAST5 = "fast5"

UPLOAD_MODES = [
    MODE_DEFAULT,
    MODE_ASSEMBLIES,
    MODE_FAST5
]

# Timeout values for sequence file data upload
# Wait at least 1 second for each mb of data
TIMEOUT_BYTES_TO_MB_DIVISOR = 1024 * 1024
# 20 minute minimum timeout
TIMEOUT_MINIMUM = 1200

SESSION_HEADERS = {
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) "
                  "Chrome/71.1.2222.33 Safari/537.36",
    "Accept-Encoding": "*",
    "Connection": "keep-alive"
}

JSON_HEADERS = {"headers": {'Content-Type': 'application/json', **SESSION_HEADERS}}

MINIMUM_IRIDA_VERSION = "23.01"


class ApiCalls(object):

    def __init__(self, client_id, client_secret,
                 base_url, username, password, timeout_multiplier=10, max_wait_time=20,
                 http_max_retries=5, http_backoff_factor=0):
        """
        Create OAuth2Session and store it
        Raises IridaConnectionError with description of error if unable to connect

        arguments:
            client_id -- client_id for creating access token.
            client_secret -- client_secret for creating access token.
            base_url -- url of the IRIDA server
            username -- username for server
            password -- password for given username
            timeout_multiplier -- number of seconds to give per MB of data being transferred

        return ApiCalls object
        """

        self.client_id = client_id
        self.client_secret = client_secret
        self.base_url = base_url
        self.username = username
        self.password = password
        self.timeout_multiplier = timeout_multiplier
        self.max_wait_time = max_wait_time
        self.http_max_retries = http_max_retries
        self.http_backoff_factor = http_backoff_factor

        self._session_lock = threading.Lock()
        self._session_set_externally = False
        self._create_session()
        self.cached_projects = None
        self.cached_samples = {}

        # these two are used when sending signals to the progress module
        self._current_upload_project_id = None
        self._current_upload_sample_name = None

        # init irida version and check if version is compatible
        self._irida_version = None
        if not self._is_irida_version_compatible(self.get_irida_version(), MINIMUM_IRIDA_VERSION):
            raise exceptions.IridaConnectionError(
                f"This API requires minimum IRIDA Version '{MINIMUM_IRIDA_VERSION}'. "
                f"IRIDA Version '{self._irida_version}' is outdated, please contact your system administrator."
            )

    @property
    def _session(self):
        try:  # Todo: rework this code without the try/catch/finally and odd exception raise
            self._session_lock.acquire()
            response = self._session_instance.options(self.base_url)
            if response.status_code != HTTPStatus.OK:
                raise Exception
            else:
                logging.debug("Existing session still works, going to reuse it.")
        except Exception:
            logging.debug("Token is probably expired, going to get a new session.")
            self._reinitialize_session()
        finally:
            self._session_lock.release()

        return self._session_instance

    def _reinitialize_session(self):
        oauth_service = self._get_oauth_service()
        access_token = self._get_access_token(oauth_service)
        _sess = oauth_service.get_session(access_token)
        # We add a HTTPAdapter with max retries so we don't fail out if one request gets lost
        logging.info("Session configured with {} retries and {} backoff factor.".format(
            self.http_max_retries, self.http_backoff_factor))
        # Backoff calculation via Retry
        # {backoff factor} * (2 ** ({number of total retries} - 1))
        # example in seconds
        # backoff = 1 = [0.5, 1, 2, 4, 8, 16, 32, 64, 128, 256, ...]
        # backoff = 2 = [1, 2, 4, 8, 16, 32, 64, 128, 256, 512, ...]
        # backoff = 10 = [5, 10, 20, 40, 80, 160, 320, 640, 1280, 2560, ...]
        retry_strategy = Retry(
            total=self.http_max_retries,
            backoff_factor=self.http_backoff_factor,
            status_forcelist=[429, 500, 502, 503, 504],
            allowed_methods=["HEAD", "GET", "PUT", "POST"],  # by default POST is excluded
        )
        # override retries built in max backoff value
        Retry.DEFAULT_BACKOFF_MAX = self.http_backoff_factor
        _sess.mount('https://', HTTPAdapter(max_retries=retry_strategy))
        _sess.mount('http://', HTTPAdapter(max_retries=retry_strategy))
        self._session_instance = _sess

    def _create_session(self):
        """
        create session to be re-used until expiry for get and post calls

        returns session (OAuth2Session object)
        """

        def validate_url_form(url):
            """
                offline 'validation' of url. parse through url and see if its malformed
            """
            parsed = urlparse(url)
            valid = len(parsed.scheme) > 0
            return valid

        if self.base_url[-1:] != "/":
            self.base_url = self.base_url + "/"
        if not validate_url_form(self.base_url):
            logging.error("Cannot create session. {} is not a valid URL".format(self.base_url))
            raise exceptions.IridaConnectionError("Cannot create session." + self.base_url + " is not a valid URL")

        self._reinitialize_session()

    def _get_oauth_service(self):
        """
        get oauth service to be used to get access token

        returns oauthService
        """

        access_token_url = urljoin(self.base_url, "oauth/token")
        oauth_service = OAuth2Service(
            client_id=self.client_id,
            client_secret=self.client_secret,
            name="irida",
            access_token_url=access_token_url,
            base_url=self.base_url
        )

        return oauth_service

    def _get_access_token(self, oauth_service):
        """
        get access token to be used to get session from oauth_service

        arguments:
            oauth_service -- O2AuthService from get_oauth_service

        returns access token
        """

        def token_decoder(return_dict):
            """
            safely parse given dictionary

            arguments:
                return_dict -- access token dictionary

            returns evaluated dictionary
            """
            # It is supposedly safer to decode bytes to string and then use ast.literal_eval than just use eval()
            try:
                irida_dict = ast.literal_eval(return_dict.decode("utf-8"))
            except (SyntaxError, ValueError):
                # SyntaxError happens when something that looks nothing like a token is returned (ex: 404 page)
                # ValueError happens with the path returns something that looks like a token, but is invalid
                #   (ex: forgetting the /api/ part of the url)
                raise ConnectionError("Unexpected response from server, URL may be incorrect")
            return irida_dict

        params = {
            "data": {
                "grant_type": "password",
                "client_id": self.client_id,
                "client_secret": self.client_secret,
                "username": self.username,
                "password": self.password
            }
        }

        try:
            access_token = oauth_service.get_access_token(
                decoder=token_decoder, **params)
        except ConnectionError as e:
            logging.error("Can not connect to IRIDA")
            raise exceptions.IridaConnectionError("Could not connect to the IRIDA server. URL may be incorrect."
                                                  " IRIDA returned with error message: {}".format(e.args))
        except KeyError as e:
            logging.error("Can not get access token from IRIDA")
            raise exceptions.IridaConnectionError("Could not get access token from IRIDA. Credentials may be incorrect."
                                                  " IRIDA returned with error message: {}".format(e.args))

        return access_token

    @staticmethod
    def _handle_rest_exception(url, e):
        """
        Given an exception, generates a more detailed description and a IRIDA specific error
        :param url: original url called (to be included in response)
        :param e: exception that was raised
        :return: IridaConnectionError to be raised by calling function
        """
        if type(e) is URLError:
            logging.error("Could not connect to IRIDA, URL '{}' responded with: {}"
                          "".format(url, str(e)))
            return exceptions.IridaConnectionError("Could not connect to IRIDA, URL '{}' responded with: {}"
                                                   "".format(url, str(e)))
        if type(e) is ConnectionError:
            # This could be anything from disconnection during post to IRIDA crashing
            logging.error("ConnectionError occurred while transferring data: " + str(e))
            raise exceptions.IridaConnectionError(e)
        else:
            logging.error("Could not connect to IRIDA, non URLError Exception occurred. URL '{}' Error: {}"
                          "".format(url, str(e)))
            return exceptions.IridaConnectionError("Could not connect to IRIDA, non URLError Exception occurred. "
                                                   "URL '{}' Error: {}".format(url, str(e)))

    @staticmethod
    def _handle_irida_exception(response):
        """
        Generates and returns an appropriate exception based on the status code in the response returned by irida

        :param response: response from irida (Non 201)
        :return: Exception Sub Class
        """
        logging.debug("response.status_code: " + str(response.status_code))
        logging.debug("response.text: " + response.text)
        if response.status_code == HTTPStatus.BAD_REQUEST:  # 400
            e = exceptions.IridaResourceError("Request to IRIDA was invalid: "
                                              "{status_code}: {err_msg}\n".format(
                                                  status_code=str(response.status_code),
                                                  err_msg=response.reason))
        elif response.status_code in [HTTPStatus.UNAUTHORIZED,
                                      HTTPStatus.FORBIDDEN]:  # 401, 403
            e = exceptions.IridaResourceError("Request to IRIDA is Forbidden. Do you have access to this resource?: "
                                              "{status_code}: {err_msg}\n".format(
                                                  status_code=str(response.status_code),
                                                  err_msg=response.reason))
        elif response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR:  # 500
            e = exceptions.IridaConnectionError("IRIDA encountered an error while handling your request: "
                                                "{status_code}: {err_msg}\n".format(
                                                    status_code=str(response.status_code),
                                                    err_msg=response.reason))
        elif response.status_code == HTTPStatus.NOT_FOUND:  # 404
            e = exceptions.IridaResourceError("The url you specified could not be reached. Please verify your URL and "
                                              "the projects/samples/etc you specified.")
        else:
            e = exceptions.IridaConnectionError("Error encountered while communicating with IRIDA: "
                                                "{status_code}: {err_msg}\n".format(
                                                    status_code=str(response.status_code),
                                                    err_msg=response.reason))
        return e

    @staticmethod
    def _is_irida_version_compatible(irida_version, minimum_irida_version):
        """
        Strips extra data from fetched irida version and checks it against set minimum compatible irida version
        param irida_version: String: actual irida version
        param minimum_irida_version: String: minimum compatible irida version
        returns: boolean: if api version is compatible with irida it is connected to
        """
        def _compare_version(v1: str, v2: str) -> int:
            """
            compares version strings
            returns: 1 if v1 > v2, 0 if v1 == v2, -1 if v1 < v2
            """
            v1, v2 = list(map(int, v1.split('.'))), list(map(int, v2.split('.')))
            for rev1, rev2 in zip_longest(v1, v2, fillvalue=0):
                if rev1 == rev2:
                    continue
                return -1 if rev1 < rev2 else 1
            return 0

        # Strip "irida-" and "-SNAPSHOT" if they are on the version id
        irida_version_stripped = irida_version.replace('-SNAPSHOT', '').replace('irida-', '')
        logging.info("Minimum Compatible IRIDA Version: " + str(minimum_irida_version))
        logging.info("Current IRIDA Version: " + str(irida_version_stripped))
        return _compare_version(irida_version_stripped, minimum_irida_version) >= 0

    def get_irida_version(self):
        """
        API call to api/version

        returns: string with version information
        """
        if self._irida_version is None:
            logging.debug("Fetching IRIDA version")

            url = f"{self.base_url}/version"
            try:
                response = self._session.get(url)
            except Exception as e:
                raise ApiCalls._handle_rest_exception(url, e)

            if response.status_code == HTTPStatus.OK:  # 200
                result = response.json()['version']
            else:
                raise self._handle_irida_exception(response)

            self._irida_version = result

        return self._irida_version

    def get_projects(self):
        """
        API call to api/projects to get list of projects

        returns list containing projects. each project is Project object.
        """

        logging.info("Loading projects.")

        if self.cached_projects is None:
            logging.debug("Loading projects from IRIDA server.")
            url = f"{self.base_url}projects"

            try:
                response = self._session.get(url)
            except Exception as e:
                raise ApiCalls._handle_rest_exception(url, e)

            if response.status_code == HTTPStatus.OK:  # 200
                result = response.json()["resource"]["resources"]
            else:
                raise self._handle_irida_exception(response)

            try:
                project_list = [
                    model.Project(
                        name=project_dict["name"],
                        description=project_dict["projectDescription"],
                        id=project_dict["identifier"]
                    )
                    for project_dict in result
                ]

            except KeyError as e:
                e.args = map(str, e.args)
                msg_arg = " ".join(e.args)
                logging.debug(msg_arg + " not found. Available keys: "
                              ", ".join(result[0].keys()))
                raise exceptions.IridaResourceError(msg_arg + " not found in data provided by IRIDA. Available keys: "
                                                    ", ".join(result[0].keys()))
            self.cached_projects = project_list
        else:
            logging.debug("Loading projects from cache.")

        return self.cached_projects

    """
    Potential future functionality:
    get_sample(sample_id): returns a single sample based on the irida sample identifier
    """

    def get_samples(self, project_id):
        """
        API call to api/projects/project_id/samples

        arguments:
            project_id -- project identifier from irida

        returns list of samples for the given project.
            each sample is a Sample object.
        """

        logging.info("Getting samples from project '{}'".format(project_id))

        if project_id not in self.cached_samples:
            url = f"{self.base_url}projects/{project_id}/samples"

            try:
                response = self._session.get(url)
            except Exception as e:
                raise ApiCalls._handle_rest_exception(url, e)

            if response.status_code != HTTPStatus.OK:  # 200
                logging.error("Encountered error while getting samples: {} {}"
                              "".format(response.status_code, response.reason))
                raise ApiCalls._handle_irida_exception(response)

            result = response.json()["resource"]["resources"]

            sample_list = []
            for sample_dict in result:
                sample_list.append(ApiCalls._build_sample_obj_from_resource(sample_dict))
            self.cached_samples[project_id] = sample_list

        return self.cached_samples[project_id]

    def get_sequence_files(self, project_id, sample_name):
        """
        API call to api/projects/project_id/sample_id/sequenceFiles
        We fetch the sample file through the project id on this route

        arguments:

            sample_name -- the sample name identifier to get from irida, relative to a project
            project_id -- the id of the project the sample is on

        returns list of sequencefile dictionary for given sample_id
        """
        # Need the sample id for upload, not the sample name.
        sample_id = self.get_sample_id(sample_name, project_id)
        return self.get_sequence_files_by_id(project_id, sample_id)

    def get_sequence_files_by_id(self, project_id, sample_id):
        """
        API call to api/projects/project_id/sample_id/sequenceFiles
        We fetch the sample file through the project id on this route

        arguments:

            sample_id -- the sample id to get from irida, relative to a project
            project_id -- the id of the project the sample is on

        returns list of sequencefile dictionary for given sample_id
        """

        logging.info("Getting sequence files from sample id '{}' on project '{}'".format(sample_id, project_id))

        url = f"{self.base_url}samples/{sample_id}/sequenceFiles"

        try:
            response = self._session.get(url)
        except Exception as e:
            raise ApiCalls._handle_rest_exception(url, e)

        if response.status_code == HTTPStatus.OK:  # 200
            # todo future development
            # This response should be parsed into SequenceFile objects
            # This is a bit tricky because Forward and Reverse reads are different files in the returned resources
            result = response.json()["resource"]["resources"]
        else:
            raise self._handle_irida_exception(response)

        return result

    def get_assemblies_files(self, sample_id):
        """
        API call to api/samples/sample_id/assemblies

        :param sample_id: sample the assemblies are on
        :return: list of assemblies files dictionary for given sample_name
        """

        logging.info("Getting assemblies files from sample '{}'".format(sample_id))

        url = f"{self.base_url}samples/{sample_id}/assemblies"

        try:
            response = self._session.get(url)
        except Exception as e:
            raise ApiCalls._handle_rest_exception(url, e)

        if response.status_code == HTTPStatus.OK:  # 200
            # todo future development if needed one day
            # This response should be returned as some sort of file object
            # This is related to how we return get_sequence_files too, but there is no real use for it at the moment,
            # yagni
            result = response.json()["resource"]["resources"]
        else:
            raise self._handle_irida_exception(response)

        return result

    def get_fast5_files(self, project_id, sample_name):
        """
        API call to api/projects/project_id/sample_id/sequenceFiles/fast5
        We fetch the fast5 files through the project id on this route

        arguments:

            sample_name -- the sample name identifier to get from irida, relative to a project
            project_id -- the id of the project the sample is on

        returns list of fast5 files dictionary for given sample_id
        """

        logging.info("Getting fast5 files from sample '{}' on project '{}'".format(sample_name, project_id))

        # Need the sample id for upload, not the sample name.
        # This is cached so it's only 1 call per project uploading to
        # TODO: In the future, this function should accept sample_id instead of sample_name
        sample_id = self.get_sample_id(sample_name, project_id)
        url = f"{self.base_url}samples/{sample_id}/fast5"

        try:
            response = self._session.get(url)
        except Exception as e:
            raise ApiCalls._handle_rest_exception(url, e)

        if response.status_code == HTTPStatus.OK:  # 200
            # todo future development if needed one day
            # This response should be returned as some sort of file object
            # This is related to how we return get_sequence_files too, but there is no real use for it at the moment,
            # yagni
            result = response.json()["resource"]["resources"]
        else:
            raise self._handle_irida_exception(response)

        return result

    def get_metadata(self, sample_id):
        """
        API call to api/samples/{sampleId}/metadata
        arguments:
            sample_id
        returns list of metadata associated with sampleID
        """

        logging.info("Getting metadata from sample id '{}'".format(sample_id))

        url = f"{self.base_url}samples/{sample_id}/metadata"

        try:
            response = self._session.get(url)
        except Exception as e:
            raise ApiCalls._handle_rest_exception(url, e)

        if response.status_code == HTTPStatus.OK:  # 200
            result = response.json()["resource"]["metadata"]
        else:
            raise self._handle_irida_exception(response)

        # TODO: api refactor project: build this data into a model/Metadata object
        return result

    def send_project(self, project, clear_cache=True):
        """
        post request to send a project to IRIDA via API
        the project being sent requires a name that is at least
            5 characters long

        arguments:
            project -- a Project object to be sent.

        returns a dictionary containing the result of post request.
        when post is successful the dictionary it returns will contain the same
            name and projectDescription that was originally sent as well as
            additional keys like createdDate and identifier.
        when post fails then an error will be raised so return statement is
            not even reached.
        """
        logging.info("Sending project to IRIDA.")

        if clear_cache:
            self.cached_projects = None
        url = f"{self.base_url}projects"
        json_obj = json.dumps(project.get_uploadable_dict())

        try:
            response = self._session.post(url, json_obj, **JSON_HEADERS)
        except Exception as e:
            raise ApiCalls._handle_rest_exception(url, e)

        if response.status_code == HTTPStatus.CREATED:  # 201
            json_res = json.loads(response.text)
        else:
            logging.error("Error sending project: {} {}".format(response.status_code, response.text))
            raise self._handle_irida_exception(response)

        return json_res

    # TODO: in the graphql api rewrite, it would be nice if these types of functions returned the Object the created
    def send_sample(self, sample, project_id):
        """
        Post request to send a sample to a project

        :param sample: Sample object to send
        :param project_id: id of project to send sample too
        :return: json response from server
        """

        logging.info("Creating sample '{}' for project '{}' on IRIDA.".format(sample.sample_name, project_id))

        self.cached_samples = {}  # reset the cache, we're updating stuff
        self.cached_projects = None

        url = f"{self.base_url}projects/{project_id}/samples"
        json_obj = json.dumps(sample.get_uploadable_dict())

        try:
            response = self._session.post(url, json_obj, **JSON_HEADERS)
        except Exception as e:
            raise ApiCalls._handle_rest_exception(url, e)

        if response.status_code == HTTPStatus.CREATED:  # 201
            json_res = json.loads(response.text)
        else:
            logging.error("Did not create sample on server. Response code is '{}' and error message is '{}'"
                          "".format(response.status_code, response.text))
            raise self._handle_irida_exception(response)

        return json_res

    def get_sample_by_id(self, sample_id):
        """
        Given a sample id, returns response from server for the baseurl/samples/sample_id endpoint
        :param sample_id:
        :return: Sample obj or None
        """
        logging.info("Getting sample info for sample id '{}'".format(sample_id))

        url = f"{self.base_url}samples/{sample_id}"

        try:
            response = self._session.get(url)
        except Exception as e:
            raise ApiCalls._handle_rest_exception(url, e)

        if response.status_code == HTTPStatus.OK:  # 200, return sample object
            sample_dict = response.json()["resource"]
            return ApiCalls._build_sample_obj_from_resource(sample_dict)
        elif response.status_code == HTTPStatus.NOT_FOUND:  # 404, return None
            return None
        else:  # any other exception
            raise self._handle_irida_exception(response)

    def get_sample_by_name(self, project_id, sample_name):
        """
        Given a project id and sample name, returns a Sample object, or None is sample does not exist
        :param project_id:
        :param sample_name:
        :return: Sample obj or None
        """

        logging.info("Getting Sample object for project id '{}' and sample name '{}'".format(project_id, sample_name))

        url = f"{self.base_url}projects/{project_id}/samples/bySampleName"
        params = {'sampleName': sample_name}

        try:
            response = self._session.get(url, params=params)
        except Exception as e:
            raise ApiCalls._handle_rest_exception(url, e)
        if response.status_code == HTTPStatus.OK:  # 200, return sample object
            logging.debug("sample found")
            sample_dict = response.json()["resource"]
            return ApiCalls._build_sample_obj_from_resource(sample_dict)
        elif response.status_code == HTTPStatus.NOT_FOUND:  # 404, return None
            logging.debug("sample not found")
            return None
        else:  # any other exception
            logging.debug("exception occurred: {}".format(response))
            raise self._handle_irida_exception(response)

    # TODO: in the graphql api rewrite these type of functions should exist in their own files per object type
    @staticmethod
    def _build_sample_obj_from_resource(resource_dict):
        """
        Given a resource dictionary for a sample, return a Sample object
        """
        # use name and description from dictionary as base parameters when creating sample
        s_name = resource_dict['sampleName']
        s_desc = resource_dict['description']
        s_id = int(resource_dict['identifier'])
        # remove them from the dict so we don't have useless duplicate data
        del resource_dict['sampleName']
        del resource_dict['description']
        del resource_dict['identifier']
        sample_obj = model.Sample(
            sample_name=s_name,
            description=s_desc,
            samp_dict=resource_dict,
            sample_id=s_id
        )
        return sample_obj

    def send_sequence_files(self, sequence_file, sample_name, project_id, upload_id, upload_mode=MODE_DEFAULT):
        """
        post request to send sequence files found in given sample argument
        raises error if either project ID or sample ID found in Sample object
        doesn't exist in irida

        arguments:
            sequence_file -- SequenceFile object to send
            sample_name -- irida sample name identifier to send to
            project_id -- irida project identifier
            upload_id -- the run to upload the files to
            upload_mode -- default:MODE_DEFAULT -- which upload mode will be used

        returns result of post request.
        """

        # update which files are being sent
        self._current_upload_project_id = project_id
        self._current_upload_sample_name = sample_name

        # Get the url's needed to send sequence files
        # Need the sample id for upload, not the sample name.
        # This is cached so it's only 1 call per project uploading to
        # TODO: In the future, this function should accept sample_id instead of sample_name
        sample_id = self.get_sample_id(sample_name, project_id)
        sample_url = f"{self.base_url}samples/{sample_id}"
        url = ApiCalls._get_sample_upload_url(sequence_file, sample_url, upload_mode)

        # Get the data encoder
        data_pkg = self._get_sequence_data_pkg(sequence_file, upload_id)
        # Generate headers from the data encoder
        headers_pkg = {'Content-Type': data_pkg.content_type, **SESSION_HEADERS}

        logging.debug("Sending files to [{}]".format(url))
        logging.debug("headers: " + str(headers_pkg))

        timeout = self._get_sequence_file_timeout(sequence_file)

        try:
            response = self._session.post(url, data=data_pkg, headers=headers_pkg, timeout=timeout)
        except Exception as e:
            logging.error("ConnectionError occurred while transferring data: " + str(e))
            raise ApiCalls._handle_rest_exception(url, e)

        if response.status_code == HTTPStatus.CREATED:
            json_res = json.loads(response.text)
        else:
            logging.error("Error while uploading [{}]: [{}]".format(sample_name, response.reason))
            raise self._handle_irida_exception(response)

        return json_res

    @staticmethod
    def _get_sample_upload_url(sequence_file, sample_url, upload_mode):
        """
        Gets the appropriate url for single end, paired end, or assemblies files.
        :param sequence_file: Sequence Fle to upload
        :param sample_url: Sample Url to upload to
        :param upload_mode: String indicating upload mode
        :return:
        """
        if upload_mode == MODE_ASSEMBLIES:
            url = f"{sample_url}/assemblies"
        elif upload_mode == MODE_FAST5:
            url = f"{sample_url}/fast5"
        elif upload_mode == MODE_DEFAULT:
            if sequence_file.is_paired_end():
                logging.debug("api_calls: sending paired-end file")
                url = f"{sample_url}/pairs"
            else:
                logging.debug("api_calls: sending single-end file")
                url = f"{sample_url}/sequenceFiles"
        else:
            error = "Upload mode '{}' is invalid. Upload mode must be one of {}".format(
                upload_mode,
                UPLOAD_MODES
            )
            logging.error(error)
            raise exceptions.IridaResourceError(error, upload_mode)

        return url

    def _get_sequence_file_timeout(self, sequence_file):
        """
        Approximates transfer time and generates a timeout according to variables defined in this module.

        These values can be overridden when importing the module
        :param sequence_file:
        :return:
        """
        # Get approximation for amount of data to send
        filesize_bytes = Path(sequence_file.file_list[0]).stat().st_size
        if sequence_file.is_paired_end():
            filesize_bytes = filesize_bytes * 2
        # Gives timeout_multiplier seconds per mb of data to transfer
        timeout_mb = (filesize_bytes * self.timeout_multiplier / TIMEOUT_BYTES_TO_MB_DIVISOR)
        # minimum time should be 20 minutes
        return timeout_mb if timeout_mb > TIMEOUT_MINIMUM else TIMEOUT_MINIMUM

    def send_metadata(self, metadata, sample_id):
        """
        Put request to add metadata to specific sample ID

        :param metadata: Metadata object
        :param sample_id: id of sample to add metadata to
        :return: json response from server
        """

        logging.info("Adding metadata to sample '{}' ".format(sample_id))

        url = f"{self.base_url}samples/{sample_id}/metadata"

        json_obj = json.dumps(metadata.get_uploadable_dict())

        try:
            response = self._session.put(url, data=json_obj, **JSON_HEADERS)
        except Exception as e:
            raise ApiCalls._handle_rest_exception(url, e)

        if response.status_code == HTTPStatus.OK:  # 200
            json_res = json.loads(response.text)
        else:
            logging.error("Did not add metadata to sample. Response code is '{}' and error message is '{}'"
                          "".format(response.status_code, response.text))
            raise self._handle_irida_exception(response)

        return json_res

    def _send_file_callback(self, monitor):
        """
        Sends data to the progress module to update file percentages
        """
        progress_percent = round(monitor.bytes_read / monitor.len * 100, 2)
        progress.send_progress(progress.ProgressData(
            sample=self._current_upload_sample_name,
            project=self._current_upload_project_id,
            progress=progress_percent
        ))
        print("Progress: ", progress_percent, "% Uploaded     \r", end="")

    def _get_sequence_data_pkg(self, sequence_file, upload_id):
        """
        Creates the data encoder, and attaches a monitor for callback functionality
        """
        # build data encoder
        encoder = ApiCalls._get_multipart_encoder(sequence_file, upload_id)
        # create callback monitor for file progress
        monitor = MultipartEncoderMonitor(encoder, self._send_file_callback)
        # override max byte read size
        # This lambda overrides httplibs hard coded 8192 byte read size
        # More details: https://github.com/requests/toolbelt/issues/75#issuecomment-237189952
        # Update: although configuring blocksize has been added to Python 3.7, it is not implemented in Requests, and is
        # labeled as "wontfix". So unless we want to move away from Multipart File Encoding, this lambda patch stays.
        # If we do move away from Multipart File Encoding, choosing a faster file transfer method that does not rely on
        # http would be better than switching to urllib3 or other python http packages.
        monitor._read = monitor.read
        monitor.read = lambda size: monitor._read(1024 * 1024)
        # return the monitor/encoder object
        return monitor

    @staticmethod
    def _get_multipart_encoder(sequence_file, upload_id):
        """
        Creates a multipart file encoder to be used for streaming files to IRIDA
        """

        logging.debug("building multipart encoder")

        boundary = "B0undary"

        if sequence_file.is_paired_end():
            # Get file names of sequence files
            file_name_a = sequence_file.file_list[0]
            file_name_b = sequence_file.file_list[1]

            file_metadata = sequence_file.properties_dict
            # miseqRunId is what irida uses to parse the upload id
            # we should think about renaming this in irida,
            # but when we do it will break compatibility with all older uploaders
            file_metadata["miseqRunId"] = str(upload_id)
            file_metadata_json = json.dumps(file_metadata)

            m_encoder = MultipartEncoder(
                fields={
                    'file1': (file_name_a.replace("\\", "/"), open(file_name_a, 'rb')),
                    'file2': (file_name_b.replace("\\", "/"), open(file_name_b, 'rb')),
                    'parameters1': (None, str(file_metadata_json), 'application/json'),
                    'parameters2': (None, str(file_metadata_json), 'application/json')
                },
                boundary=boundary
            )

            return m_encoder
        else:
            file_name = sequence_file.file_list[0]

            file_metadata = sequence_file.properties_dict
            # miseqRunId is what irida uses to parse the upload id
            # we should think about renaming this in irida,
            # but when we do it will break compatibility with all older uploaders
            file_metadata["miseqRunId"] = str(upload_id)
            file_metadata_json = json.dumps(file_metadata)

            m_encoder = MultipartEncoder(
                fields={
                    'file': (file_name.replace("\\", "/"), open(file_name, 'rb')),
                    'parameters': (None, str(file_metadata_json), 'application/json'),
                },
                boundary=boundary
            )

            return m_encoder

    def create_seq_run(self, metadata, sequencing_run_type):
        """
        Create a sequencing run.

        uploadStatus "UPLOADING"

        There are some parsed metadata keys from the SampleSheet.csv that are
        currently not accepted/used by the API so they are discarded.
        Everything not in the acceptable_properties list below is discarded.

        arguments:
            metadata -- SequencingRun's metadata
            sequencing_run_type -- string: used as the identifier for the type of sequencing run being uploaded

        returns: the sequencing run identifier for the sequencing run that was created
        """

        logging.debug("Creating new sequencing run on IRIDA")

        metadata_dict = metadata.copy()
        # metadata_dict requires the workflow parameter or else IRIDA will not create the seq run
        if 'workflow' not in metadata_dict:
            metadata_dict['workflow'] = 'workflow'

        url = f"{self.base_url}sequencingrun/{sequencing_run_type}"

        acceptable_properties = [
            "layoutType", "chemistry", "projectName",
            "experimentName", "application", "uploadStatus",
            "investigatorName", "createdDate", "assay", "description",
            "workflow", "readLengths"]

        metadata_dict["uploadStatus"] = "UPLOADING"

        keys_to_remove = []
        for key in metadata_dict.keys():
            if key not in acceptable_properties:
                keys_to_remove.append(key)

        for key in keys_to_remove:
            del metadata_dict[key]

        json_obj = json.dumps(metadata_dict)

        try:
            response = self._session.post(url, json_obj, **JSON_HEADERS)
        except Exception as e:
            raise ApiCalls._handle_rest_exception(url, e)

        if response.status_code == HTTPStatus.CREATED:  # 201
            json_res = json.loads(response.text)
        else:
            logging.error("Encountered error while creating sequence run: {} {}"
                          "".format(response.status_code, response.reason))
            raise self._handle_irida_exception(response)

        # Grab the run identifier from the returned json
        sequencing_run_id = json_res['resource']['identifier']
        logging.debug("Sequencing run id '{}' has been created".format(sequencing_run_id))
        return sequencing_run_id

    def get_seq_runs(self):
        """
        Get list of all SequencingRun objects

        return list of SequencingRuns
        """
        logging.debug("Getting sequencing runs")

        url = f"{self.base_url}sequencingrun"

        try:
            response = self._session.get(url)
        except Exception as e:
            raise ApiCalls._handle_rest_exception(url, e)

        if response.status_code == HTTPStatus.OK:  # 200
            json_res_list = response.json()["resource"]["resources"]
        else:
            logging.error("Encountered error while getting sequence runs: {} {}"
                          "".format(response.status_code, response.reason))
            raise self._handle_irida_exception(response)

        return json_res_list

    def set_seq_run_complete(self, identifier):

        """
        Update a sequencing run's upload status to "COMPLETE"

        arguments:
            identifier -- the id of the sequencing run to be updated

        returns result of patch request
        """

        status = "COMPLETE"
        json_res = self._set_seq_run_upload_status(identifier, status)

        return json_res

    def set_seq_run_uploading(self, identifier):

        """
        Update a sequencing run's upload status to "UPLOADING"

        arguments:
            identifier -- the id of the sequencing run to be updated

        returns result of patch request
        """

        status = "UPLOADING"
        json_res = self._set_seq_run_upload_status(identifier, status)

        return json_res

    def set_seq_run_error(self, identifier):

        """
        Update a sequencing run's upload status to "ERROR".

        arguments:
            identifier -- the id of the sequencing run to be updated

        returns result of patch request
        """

        status = "ERROR"
        json_res = self._set_seq_run_upload_status(identifier, status)

        return json_res

    def _set_seq_run_upload_status(self, identifier, status):

        """
        Update a sequencing run's upload status to the given status argument

        arguments:
            identifier -- the id of the sequencing run to be updated
            status     -- string that the sequencing run will be updated
                          with

        returns result of patch request
        """

        logging.debug("Setting sequencing run '{}' to '{}'".format(identifier, status))

        url = f"{self.base_url}sequencingrun/{identifier}"
        update_dict = {"uploadStatus": status}
        json_obj = json.dumps(update_dict)

        try:
            response = self._session.patch(url, json_obj, **JSON_HEADERS)
        except Exception as e:
            raise ApiCalls._handle_rest_exception(url, e)

        if response.status_code == HTTPStatus.OK:  # 200
            json_res = json.loads(response.text)
        else:
            logging.error("Encountered error while changing upload status of run '{}' to '{}': {} {}"
                          "".format(identifier, status, response.status_code, response.reason))
            raise exceptions.IridaConnectionError("Error: " + str(response.status_code) + " " + response.reason)

        return json_res

    def project_exists(self, project_id):
        """
        Check if a project exists

        :param project_id: project that we are checking for existence
        :return: True or False
        """
        logging.debug("project exists: {}".format(project_id))
        project_id = str(project_id)
        project_list = self.get_projects()
        return any([p.id == project_id for p in project_list])

    def try_project_access(self, project_id):
        """
        Attempts a get request to a project, and returns the HTTPStatus that IRIDA returns
        This is useful for determining if a user does not have access to a project or if it does not exist at all.

        :param project_id:
        :return: HTTPStatus(IntEnum)
        """
        logging.debug("Trying to access project: {}".format(project_id))

        url = f"{self.base_url}projects/{project_id}"

        try:
            response = self._session.get(url)
        except Exception as e:
            # Handle any exceptions where communication with the server fails
            raise ApiCalls._handle_rest_exception(url, e)

        logging.debug("IRIDA responded with status code: {}".format(response.status_code))
        return response.status_code

    # TODO: this function should be removed during the graphql api rewrite
    def sample_exists(self, sample_name, project_id):
        """
        Given a sample name and project id, returns True or False for if sample exists
        :param sample_name:
        :param project_id:
        :return:
        """
        return True if (self.get_sample_by_name(project_id, sample_name) is not None) else False

    # TODO: This function should be removed during the graphql api rewrite.
    # It does not need to exist because get_sample_by_name returns an object with this information, but it is baked into
    # other functions. Additionally we don't want to break any user scripts until we fully rewrite the api with graphql
    def get_sample_id(self, sample_name, project_id):  # todo remove this
        """
        Given a sample name and project id, returns the sample id, or False if it doesn't exist

        This method shares sample caching with the get_samples method

        :param sample_name: sample to confirm existence of
        :param project_id: project that we think the sample is on
        :return: Integer of the sample identifier if it exists, otherwise False
        """
        logging.debug("sample exists: sample: {}, on project: {}".format(sample_name, project_id))
        res = self.get_sample_by_name(project_id, sample_name)
        return res.sample_id if (res is not None) else False



"""
Test cases that check parts of the models that don't get validated through standard execution
"""

import unittest
import os
from unittest.mock import patch

from iridauploader import model

path_to_module = os.path.abspath(os.path.dirname(__file__))
if len(path_to_module) == 0:
    path_to_module = '.'


class TestProject(unittest.TestCase):
    """
    Tests for Project object
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_string_output(self):
        """
        Test the __str__ method
        """
        proj = model.Project("p1", None, "desc", 1)
        p_str = str(proj)
        self.assertEqual('ID:1 Name: p1 Description: desc', p_str)


class TestSample(unittest.TestCase):
    """
    Tests for Sample object
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_string_output(self):
        """
        Test the __str__ method
        """
        samp = model.Sample("s1", "desc", 1, None, 1)
        s_str = str(samp)
        self.assertEqual("{'sampleName': 's1', 'description': 'desc', 'sequenceFile': 'None'}", s_str)

    def test_get_dict(self):
        """
        test dictionary building
        """
        samp = model.Sample("s1", "desc", 1, None, 1)
        d = {'_description': 'desc',
             '_sample_dict': {},
             '_sample_id': 1,
             '_sample_name': 's1',
             '_sample_number': 1,
             '_sequence_file': None,
             '_skip': False}
        self.assertEqual(d, samp.get_dict())

    def test_get_sample_dict(self):
        """
        test metadata dictionary
        """
        samp = model.Sample(
            sample_name="s1",
            description="desc",
            sample_number=1,
            samp_dict={"some": "values", "are": "here"},
            sample_id=1
        )
        self.assertEqual({"some": "values", "are": "here"}, samp.sample_dict)

    def test_get_item_none(self):
        """
        test __getitem__ for item that doesn't exist
        """
        samp = model.Sample("s1", "desc", 1, None, 1)
        self.assertIsNone(samp.__getitem__("not a key"))


class TestSequenceFile(unittest.TestCase):
    """
    Tests for SequenceFile object
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_string_output(self):
        """
        Test the __str__ method
        """
        seq = model.SequenceFile(["file1", "file2"], {"some": "thing"})
        s_str = str(seq)
        self.assertEqual('{\'propertyDict\': "{\'some\': \'thing\'}", \'fileList\': "[\'file1\', ' '\'file2\']"}',
                         s_str)

    def test_get(self):
        """
        test get
        """
        seq = model.SequenceFile(["file1", "file2"], {"some": "thing"})
        self.assertEqual("thing", seq.get("some"))

    def test_get_item_none(self):
        """
        test get for item that doesn't exist
        """
        seq = model.SequenceFile(["file1", "file2"], {"some": "thing"})
        self.assertIsNone(seq.get("not a key"))

    def test_get_dict(self):
        """
        test dictionary building
        """
        seq = model.SequenceFile(["file1", "file2"], {"some": "thing"})
        d = {'_file_list': ['file1', 'file2'], '_properties_dict': {'some': 'thing'}}
        self.assertEqual(d, seq.get_dict())


class TestModelValidationError(unittest.TestCase):
    """
    Tests for ModelValidationError object
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_properties(self):
        """
        test properties
        """
        m = model.exceptions.ModelValidationError("msg", "this")
        self.assertEqual("this", m.object)
        self.assertEqual("msg", m.message)

    def test_string_output(self):
        """
        Test the __str__ method
        """
        m = model.exceptions.ModelValidationError("msg", None)
        m_str = str(m)
        self.assertEqual('msg', m_str)


class TestSequencingRun(unittest.TestCase):
    """
    Tests for SequencingRun object
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_properties(self):
        """
        test properties
        """
        s = model.SequencingRun({"some": "thing"}, [], "miseq")
        s.metadata = {"some", "where"}
        self.assertEqual({"some", "where"}, s.metadata)
        s.project_list = ["items", "here"]
        self.assertEqual(["items", "here"], s.project_list)
        s.sequencing_run_type = "miniseq"
        self.assertEqual("miniseq", s.sequencing_run_type)



"""
Metadata defines and stores metadata in (key: value) pairs using a dictionary

(key: value) pairs can take any string so long as they are structured as such:
    { "Key_1": "Value_1", "Key_2": "Value_2" ... }
"""


class Metadata:

    def __init__(self, metadata=None, project_id=None, sample_name=None):
        if metadata is None:
            metadata = {}
        self._metadata = metadata
        self._project_id = project_id
        self._sample_name = sample_name

    @property
    def metadata(self):
        return self._metadata

    @metadata.setter
    def metadata(self, metadata):
        self._metadata = metadata

    @property
    def project_id(self):
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        self._project_id = project_id

    @property
    def sample_name(self):
        return self._sample_name

    @sample_name.setter
    def sample_name(self, sample_name):
        self._sample_name = sample_name

    def add_metadata_entry(self, key, value):
        self._metadata[key] = value

    def get_uploadable_dict(self):
        uploadable_dict = {}
        for key, value in self._metadata.items():
            uploadable_dict[key] = {"type": "text", "value": "{}".format(value)}
        return uploadable_dict



import unittest
from unittest.mock import patch
from os import path

from iridauploader.core import cli

path_to_module = path.abspath(path.dirname(__file__))
if len(path_to_module) == 0:
    path_to_module = '.'


class TestCliEntryPoints(unittest.TestCase):
    """
    Tests the core.cli package entry points into core.upload via main()
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @patch("iridauploader.core.upload.upload_run_single_entry")
    @patch("iridauploader.core.cli._config_uploader")
    @patch("iridauploader.core.cli.init_argparser")
    def test_upload_regular(self, mock_init_argparser, mock_configure_uploader, mock_upload_run_single_entry):

        class StubArgs:
            """
            Contains dummy properties that mimic the actual arguments object
            """
            @property
            def directory(self):
                return path_to_module

            @property
            def force(self):
                return False

            @property
            def continue_partial(self):
                return False

            @property
            def batch(self):
                return False

            @property
            def upload_mode(self):
                return None

        stub_args_object = StubArgs()
        # stub_argparser returns the above args object
        stub_argparser = unittest.mock.MagicMock()
        stub_argparser.parse_args.side_effect = [stub_args_object]
        # mock_init_argparser returns the above stub_argparser, which contains the args object
        mock_init_argparser.side_effect = [stub_argparser]

        mock_configure_uploader.side_effect = [None]
        # mock the core code exiting
        stub_upload_exit = unittest.mock.MagicMock()
        stub_upload_exit.exit_code.side_effect = 0
        mock_upload_run_single_entry.side_effect = [stub_upload_exit]

        cli.main()

        # ensure the configuration is called
        mock_configure_uploader.assert_called_once_with(stub_args_object)
        # Verify final call
        mock_upload_run_single_entry.assert_called_once_with(stub_args_object.directory,
                                                             stub_args_object.force,
                                                             stub_args_object.upload_mode,
                                                             stub_args_object.continue_partial)

    @patch("iridauploader.core.upload.batch_upload_single_entry")
    @patch("iridauploader.core.cli._config_uploader")
    @patch("iridauploader.core.cli.init_argparser")
    def test_upload_batch(self, mock_init_argparser, mock_configure_uploader, mock_batch_upload_single_entry):

        class StubArgs:
            """
            Contains dummy properties that mimic the actual arguments object
            """
            @property
            def directory(self):
                return path_to_module

            @property
            def force(self):
                return False

            @property
            def continue_partial(self):
                return False

            @property
            def batch(self):
                return True

            @property
            def upload_mode(self):
                return None

        stub_args_object = StubArgs()
        # stub_argparser returns the above args object
        stub_argparser = unittest.mock.MagicMock()
        stub_argparser.parse_args.side_effect = [stub_args_object]
        # mock_init_argparser returns the above stub_argparser, which contains the args object
        mock_init_argparser.side_effect = [stub_argparser]

        mock_configure_uploader.side_effect = [None]
        # mock the core code exiting
        stub_upload_exit = unittest.mock.MagicMock()
        stub_upload_exit.exit_code.side_effect = 0
        mock_batch_upload_single_entry.side_effect = [stub_upload_exit]

        cli.main()

        # ensure the configuration is called
        mock_configure_uploader.assert_called_once_with(stub_args_object)
        # Verify final call
        mock_batch_upload_single_entry.assert_called_once_with(stub_args_object.directory,
                                                               stub_args_object.force,
                                                               stub_args_object.upload_mode,
                                                               stub_args_object.continue_partial)



import unittest
from unittest.mock import patch, call
from os import path
import os

from iridauploader.api import MODE_DEFAULT
from iridauploader.core import upload, logger, exit_return
from iridauploader.config import config
from iridauploader.model import DirectoryStatus

path_to_module = path.abspath(path.dirname(__file__))
if len(path_to_module) == 0:
    path_to_module = '.'


class TestUploadRunSingleEntry(unittest.TestCase):
    """
    Tests the core.upload.upload_run_single_entry function
    """

    # Reusable stubs
    class StubValidationResult:
        _valid = True
        _error_count = 0
        _error_list = []

        def is_valid(self):
            return self._valid

        def error_count(self):
            return self._error_count

        @property
        def error_list(self):
            return self._error_list

    class StubDirectoryStatus:
        directory = path.join(path_to_module, "fake_ngs_data")
        _status = DirectoryStatus.NEW
        _message = ""

        def status_equals(self, status):
            return status == self._status

        @property
        def message(self):
            return self._message

        @property
        def status(self):
            return self._status
        #
        # @status.setter
        # def status(self, status):
        #     self._status = status

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)
        # config.setup()
        config._init_config_parser()

    def tearDown(self):
        print("Cleaning up status file")
        log_file_path = path.join(path_to_module, "fake_ngs_data", "irida-uploader.log")
        if path.exists(log_file_path):
            os.remove(log_file_path)
        status_file_path = path.join(path_to_module, "fake_ngs_data", "irida_uploader_status.info")
        if path.exists(status_file_path):
            os.remove(status_file_path)

        # Clean up the logger in the case where a test fails to complete
        print("Cleaning up directory logger")
        if logger.directory_logger:
            logger.remove_directory_logger()

    @patch("iridauploader.core.upload.upload_helpers")
    @patch("iridauploader.core.upload._validate_and_upload")
    @patch("iridauploader.core.upload.api_handler")
    @patch("iridauploader.core.upload.parsing_handler")
    def test_invalid_directory_status(self, mock_parsing_handler, mock_api_handler,
                                      mock_validate_and_upload, mock_upload_helpers):
        """
        Checks that function exits when directory status is invalid
        :return:
        """

        stub_directory_status = self.StubDirectoryStatus()
        stub_directory_status._status = DirectoryStatus.INVALID

        mock_parsing_handler.get_run_status.side_effect = [stub_directory_status]
        mock_upload_helpers.directory_has_readonly_conflict.side_effect = [False]

        result = upload.upload_run_single_entry(stub_directory_status.directory)

        # verify result
        self.assertEqual(result.exit_code, exit_return.EXIT_CODE_ERROR)

        # verify calls occurred
        mock_parsing_handler.get_run_status.assert_called_with(stub_directory_status.directory)

        # ensure upload did not occur
        mock_validate_and_upload.assert_not_called()

    @patch("iridauploader.core.upload.upload_helpers")
    @patch("iridauploader.core.upload._validate_and_upload")
    @patch("iridauploader.core.upload.api_handler")
    @patch("iridauploader.core.upload.parsing_handler")
    def test_completed_directory_status(self, mock_parsing_handler, mock_api_handler,
                                        mock_validate_and_upload, mock_upload_helpers):
        """
        Checks that function exits when directory status is complete
        :return:
        """

        stub_directory_status = self.StubDirectoryStatus()
        stub_directory_status._status = DirectoryStatus.COMPLETE

        mock_parsing_handler.get_run_status.side_effect = [stub_directory_status]
        mock_upload_helpers.directory_has_readonly_conflict.side_effect = [False]

        result = upload.upload_run_single_entry(stub_directory_status.directory)

        # verify result
        self.assertEqual(result.exit_code, exit_return.EXIT_CODE_ERROR)

        # verify calls occurred
        mock_parsing_handler.get_run_status.assert_called_with(stub_directory_status.directory)

        # ensure upload did not occur
        mock_validate_and_upload.assert_not_called()

    @patch("iridauploader.core.upload.upload_helpers")
    @patch("iridauploader.core.upload._validate_and_upload")
    @patch("iridauploader.core.upload.api_handler")
    @patch("iridauploader.core.upload.parsing_handler")
    def test_error_directory_status(self, mock_parsing_handler, mock_api_handler,
                                    mock_validate_and_upload, mock_upload_helpers):
        """
        Checks that function exits when directory status is error
        :return:
        """

        stub_directory_status = self.StubDirectoryStatus()
        stub_directory_status._status = DirectoryStatus.ERROR

        mock_parsing_handler.get_run_status.side_effect = [stub_directory_status]
        mock_upload_helpers.directory_has_readonly_conflict.side_effect = [False]

        result = upload.upload_run_single_entry(stub_directory_status.directory)

        # verify result
        self.assertEqual(result.exit_code, exit_return.EXIT_CODE_ERROR)

        # verify calls occurred
        mock_parsing_handler.get_run_status.assert_called_with(stub_directory_status.directory)

        # ensure upload did not occur
        mock_validate_and_upload.assert_not_called()

    @patch("iridauploader.core.upload.upload_helpers")
    @patch("iridauploader.core.upload._validate_and_upload")
    @patch("iridauploader.core.upload.api_handler")
    @patch("iridauploader.core.upload.parsing_handler")
    def test_partial_directory_status(self, mock_parsing_handler, mock_api_handler,
                                      mock_validate_and_upload, mock_upload_helpers):
        """
        Checks that function exits when directory status is partial
        :return:
        """

        stub_directory_status = self.StubDirectoryStatus()
        stub_directory_status._status = DirectoryStatus.PARTIAL

        mock_parsing_handler.get_run_status.side_effect = [stub_directory_status]
        mock_upload_helpers.directory_has_readonly_conflict.side_effect = [False]

        result = upload.upload_run_single_entry(stub_directory_status.directory)

        # verify result
        self.assertEqual(result.exit_code, exit_return.EXIT_CODE_ERROR)

        # verify calls occurred
        mock_parsing_handler.get_run_status.assert_called_with(stub_directory_status.directory)

        # ensure upload
        mock_validate_and_upload.assert_not_called()

    @patch("iridauploader.core.upload.upload_helpers")
    @patch("iridauploader.core.upload._validate_and_upload")
    @patch("iridauploader.core.upload.api_handler")
    @patch("iridauploader.core.upload.parsing_handler")
    def test_completed_force_directory_status(self, mock_parsing_handler, mock_api_handler,
                                              mock_validate_and_upload, mock_upload_helpers):
        """
        Checks that function with force when directory status is complete
        :return:
        """

        stub_directory_status = self.StubDirectoryStatus()
        stub_directory_status._status = DirectoryStatus.COMPLETE

        mock_parsing_handler.get_run_status.side_effect = [stub_directory_status]
        mock_api_handler.get_default_upload_mode.side_effect = ["mock_mode"]
        mock_validate_and_upload.side_effect = ["mock_result"]
        mock_upload_helpers.directory_has_readonly_conflict.side_effect = [False]

        result = upload.upload_run_single_entry(stub_directory_status.directory, force_upload=True)

        # verify result
        self.assertEqual(result, "mock_result")

        # verify calls occurred
        mock_parsing_handler.get_run_status.assert_called_with(stub_directory_status.directory)

        # ensure upload
        mock_validate_and_upload.assert_called_with(stub_directory_status, "mock_mode", False)

    @patch("iridauploader.core.upload.upload_helpers")
    @patch("iridauploader.core.upload._validate_and_upload")
    @patch("iridauploader.core.upload.api_handler")
    @patch("iridauploader.core.upload.parsing_handler")
    def test_partial_force_directory_status(self, mock_parsing_handler, mock_api_handler,
                                            mock_validate_and_upload, mock_upload_helpers):
        """
        Checks that function continues with force when directory status is partial
        :return:
        """

        stub_directory_status = self.StubDirectoryStatus()
        stub_directory_status._status = DirectoryStatus.PARTIAL

        mock_parsing_handler.get_run_status.side_effect = [stub_directory_status]
        mock_api_handler.get_default_upload_mode.side_effect = ["mock_mode"]
        mock_validate_and_upload.side_effect = ["mock_result"]
        mock_upload_helpers.directory_has_readonly_conflict.side_effect = [False]

        result = upload.upload_run_single_entry(stub_directory_status.directory, force_upload=True)

        # verify result
        self.assertEqual(result, "mock_result")

        # verify calls occurred
        mock_parsing_handler.get_run_status.assert_called_with(stub_directory_status.directory)

        # ensure upload
        mock_validate_and_upload.assert_called_with(stub_directory_status, "mock_mode", False)

    @patch("iridauploader.core.upload.upload_helpers")
    @patch("iridauploader.core.upload._validate_and_upload")
    @patch("iridauploader.core.upload.api_handler")
    @patch("iridauploader.core.upload.parsing_handler")
    def test_partial_continue_directory_status(self, mock_parsing_handler, mock_api_handler,
                                               mock_validate_and_upload, mock_upload_helpers):
        """
        Checks that function continues with force when directory status is partial
        :return:
        """

        stub_directory_status = self.StubDirectoryStatus()
        stub_directory_status._status = DirectoryStatus.PARTIAL

        mock_parsing_handler.get_run_status.side_effect = [stub_directory_status]
        mock_api_handler.get_default_upload_mode.side_effect = ["mock_mode"]
        mock_validate_and_upload.side_effect = ["mock_result"]
        mock_upload_helpers.directory_has_readonly_conflict.side_effect = [False]

        result = upload.upload_run_single_entry(
            stub_directory_status.directory, force_upload=False, continue_upload=True)

        # verify result
        self.assertEqual(result, "mock_result")

        # verify calls occurred
        mock_parsing_handler.get_run_status.assert_called_with(stub_directory_status.directory)

        # ensure upload
        mock_validate_and_upload.assert_called_with(stub_directory_status, "mock_mode", True)

    @patch("iridauploader.core.upload.upload_helpers")
    @patch("iridauploader.core.upload._validate_and_upload")
    @patch("iridauploader.core.upload.api_handler")
    @patch("iridauploader.core.upload.parsing_handler")
    def test_error_force_directory_status(self, mock_parsing_handler, mock_api_handler,
                                          mock_validate_and_upload, mock_upload_helpers):
        """
        Checks that function continues with force when directory status is error
        :return:
        """

        stub_directory_status = self.StubDirectoryStatus()
        stub_directory_status._status = DirectoryStatus.ERROR

        mock_parsing_handler.get_run_status.side_effect = [stub_directory_status]
        mock_api_handler.get_default_upload_mode.side_effect = ["mock_mode"]
        mock_validate_and_upload.side_effect = ["mock_result"]
        mock_upload_helpers.directory_has_readonly_conflict.side_effect = [False]

        result = upload.upload_run_single_entry(stub_directory_status.directory, force_upload=True)

        # verify result
        self.assertEqual(result, "mock_result")

        # verify calls occurred
        mock_parsing_handler.get_run_status.assert_called_with(stub_directory_status.directory)

        # ensure upload
        mock_validate_and_upload.assert_called_with(stub_directory_status, "mock_mode", False)

    @patch("iridauploader.core.upload.upload_helpers")
    @patch("iridauploader.core.upload._validate_and_upload")
    @patch("iridauploader.core.upload.api_handler")
    @patch("iridauploader.core.upload.parsing_handler")
    def test_delayed_force_directory_status(self, mock_parsing_handler, mock_api_handler,
                                            mock_validate_and_upload, mock_upload_helpers):
        """
        Checks that function continues with force when directory status is delayed
        :return:
        """

        stub_directory_status = self.StubDirectoryStatus()
        stub_directory_status._status = DirectoryStatus.DELAYED

        mock_parsing_handler.get_run_status.side_effect = [stub_directory_status]
        mock_api_handler.get_default_upload_mode.side_effect = ["mock_mode"]
        mock_validate_and_upload.side_effect = ["mock_result"]
        mock_upload_helpers.directory_has_readonly_conflict.side_effect = [False]

        result = upload.upload_run_single_entry(stub_directory_status.directory, force_upload=True)

        # verify result
        self.assertEqual(result, "mock_result")

        # verify calls occurred
        mock_parsing_handler.get_run_status.assert_called_with(stub_directory_status.directory)

        # ensure upload
        mock_validate_and_upload.assert_called_with(stub_directory_status, "mock_mode", False)

    @patch("iridauploader.core.upload.upload_helpers")
    @patch("iridauploader.core.upload._validate_and_upload")
    @patch("iridauploader.core.upload.api_handler")
    @patch("iridauploader.core.upload.parsing_handler")
    def test_new_directory_status(self, mock_parsing_handler, mock_api_handler,
                                  mock_validate_and_upload, mock_upload_helpers):
        """
        Checks that function continues when directory status is new
        :return:
        """

        stub_directory_status = self.StubDirectoryStatus()
        stub_directory_status._status = DirectoryStatus.NEW

        mock_parsing_handler.get_run_status.side_effect = [stub_directory_status]
        mock_api_handler.get_default_upload_mode.side_effect = ["mock_mode"]
        mock_validate_and_upload.side_effect = ["mock_result"]
        mock_upload_helpers.directory_has_readonly_conflict.side_effect = [False]

        result = upload.upload_run_single_entry(stub_directory_status.directory)

        # verify result
        self.assertEqual(result, "mock_result")

        # verify calls occurred
        mock_parsing_handler.get_run_status.assert_called_with(stub_directory_status.directory)

        # ensure upload occurs
        mock_validate_and_upload.assert_called_with(stub_directory_status, "mock_mode", False)

    @patch("iridauploader.progress.upload_status._set_run_delayed")
    @patch("iridauploader.core.upload.upload_helpers")
    @patch("iridauploader.core.upload._validate_and_upload")
    @patch("iridauploader.core.upload.api_handler")
    @patch("iridauploader.core.upload.parsing_handler")
    def test_new_with_delay_config_directory_status(self, mock_parsing_handler, mock_api_handler,
                                                    mock_validate_and_upload, mock_upload_helpers,
                                                    mock_set_run_delayed):
        """
        Checks that function exits with success when directory status is new and there is a delay set
        :return:
        """
        # set a delay
        config.set_config_options(delay=1)

        stub_directory_status = self.StubDirectoryStatus()
        stub_directory_status._status = DirectoryStatus.NEW

        mock_parsing_handler.get_run_status.side_effect = [stub_directory_status]
        mock_set_run_delayed.side_effect = [None]
        mock_upload_helpers.directory_has_readonly_conflict.side_effect = [False]

        result = upload.upload_run_single_entry(stub_directory_status.directory)

        # verify result
        self.assertEqual(result.exit_code, exit_return.EXIT_CODE_SUCCESS)

        # verify calls occurred
        mock_parsing_handler.get_run_status.assert_called_with(stub_directory_status.directory)
        mock_set_run_delayed.assert_called_with(stub_directory_status)

        # ensure upload did not occur
        mock_validate_and_upload.assert_not_called()

    # @patch("iridauploader.progress.upload_status._delayed_time_has_passed")
    @patch("iridauploader.core.upload.upload_helpers")
    @patch("iridauploader.core.upload._validate_and_upload")
    @patch("iridauploader.core.upload.api_handler")
    @patch("iridauploader.core.upload.parsing_handler")
    def test_delay_time_has_passed_directory_status(self, mock_parsing_handler, mock_api_handler,
                                                    mock_validate_and_upload, mock_upload_helpers):
        """
        Checks that function exits with success when directory status is delayed and the delay has passed
        :return:
        """
        # set a delay for 0 as time has passed
        config.set_config_options(delay=0)

        stub_directory_status = self.StubDirectoryStatus()
        stub_directory_status._status = DirectoryStatus.DELAYED

        mock_parsing_handler.get_run_status.side_effect = [stub_directory_status]
        mock_api_handler.get_default_upload_mode.side_effect = ["mock_mode"]
        mock_validate_and_upload.side_effect = ["mock_result"]
        mock_upload_helpers.directory_has_readonly_conflict.side_effect = [False]

        result = upload.upload_run_single_entry(stub_directory_status.directory)

        # verify result
        self.assertEqual(result, "mock_result")

        # verify calls occurred
        mock_parsing_handler.get_run_status.assert_called_with(stub_directory_status.directory)

        # ensure upload occurs
        mock_validate_and_upload.assert_called_with(stub_directory_status, "mock_mode", False)

    @patch("iridauploader.progress.upload_status._delayed_time_has_passed")
    @patch("iridauploader.core.upload.upload_helpers")
    @patch("iridauploader.core.upload._validate_and_upload")
    @patch("iridauploader.core.upload.api_handler")
    @patch("iridauploader.core.upload.parsing_handler")
    def test_delay_not_passed_directory_status(self, mock_parsing_handler, mock_api_handler,
                                               mock_validate_and_upload, mock_upload_helpers,
                                               mock_delayed_time_has_passed):
        """
        Checks that function exits with success when directory status is delayed and there is still delay time
        :return:
        """
        # set a delay
        config.set_config_options(delay=1)

        stub_directory_status = self.StubDirectoryStatus()
        stub_directory_status._status = DirectoryStatus.DELAYED

        mock_parsing_handler.get_run_status.side_effect = [stub_directory_status]
        mock_delayed_time_has_passed.side_effect = [False]
        mock_upload_helpers.directory_has_readonly_conflict.side_effect = [False]

        result = upload.upload_run_single_entry(stub_directory_status.directory)

        # verify result
        self.assertEqual(result.exit_code, exit_return.EXIT_CODE_SUCCESS)

        # verify calls occurred
        mock_parsing_handler.get_run_status.assert_called_with(stub_directory_status.directory)
        mock_delayed_time_has_passed.assert_called_with(stub_directory_status, 1)

        # ensure upload did not occur
        mock_validate_and_upload.assert_not_called()

    @patch("iridauploader.core.upload.upload_helpers")
    @patch("iridauploader.core.upload._validate_and_upload")
    @patch("iridauploader.core.upload.api_handler")
    @patch("iridauploader.core.upload.parsing_handler")
    def test_directory_not_writable_no_readonly_status(self, mock_parsing_handler, mock_api_handler,
                                                       mock_validate_and_upload, mock_upload_helpers):
        """
        Checks that function exits with success when directory status is delayed and there is still delay time
        :return:
        """

        stub_directory_status = self.StubDirectoryStatus()
        stub_directory_status._status = DirectoryStatus.DELAYED

        mock_parsing_handler.get_run_status.side_effect = [stub_directory_status]
        mock_upload_helpers.directory_has_readonly_conflict.side_effect = [True]

        result = upload.upload_run_single_entry(stub_directory_status.directory)

        # verify result
        self.assertEqual(result.exit_code, exit_return.EXIT_CODE_ERROR)

        # verify calls occurred
        mock_parsing_handler.get_run_status.assert_called_with(stub_directory_status.directory)
        mock_upload_helpers.directory_has_readonly_conflict.assert_called_with(stub_directory_status.directory)

        # ensure upload did not occur
        mock_validate_and_upload.assert_not_called()


class TestBatchUploadSingleEntry(unittest.TestCase):
    """
    Tests the core.upload.batch_upload_single_entry function
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)
        config._init_config_parser()

    @patch("iridauploader.core.upload._validate_and_upload")
    @patch("iridauploader.core.upload.parsing_handler")
    def test_valid(self, mock_parsing_handler, mock_validate_and_upload):
        """
        Makes sure that _validate_and_upload is only called on a new run
        :return:
        """
        class StubDirectoryStatus:
            def __init__(self, directory, status):
                self.directory = directory
                self.status = status
                self.message = None

            def status_equals(self, other_status):
                return self.status == other_status

        # add mock data to the function calls that are essential
        stub_directory_status_valid = StubDirectoryStatus("valid", DirectoryStatus.NEW)
        stub_directory_status_invalid = StubDirectoryStatus("invalid", DirectoryStatus.INVALID)
        stub_directory_status_complete = StubDirectoryStatus("complete", DirectoryStatus.COMPLETE)
        stub_directory_status_partial = StubDirectoryStatus("partial", DirectoryStatus.PARTIAL)

        mock_parsing_handler.get_run_status_list.side_effect = [
            [stub_directory_status_valid,
             stub_directory_status_invalid,
             stub_directory_status_complete,
             stub_directory_status_partial]
        ]
        mock_parsing_handler.parse_and_validate.side_effect = ["Fake Sequencing Run"]

        # start
        upload.batch_upload_single_entry("fake_directory", upload_mode=MODE_DEFAULT)

        # validate calls only happen once
        mock_validate_and_upload.assert_called_once_with(stub_directory_status_valid, MODE_DEFAULT, False)

    @patch("iridauploader.core.upload._validate_and_upload")
    @patch("iridauploader.core.upload.parsing_handler")
    def test_valid_force(self, mock_parsing_handler, mock_validate_and_upload):
        """
        Makes sure that _validate_and_upload is called on all runs except invalid
        :return:
        """
        class StubDirectoryStatus:
            def __init__(self, directory, status):
                self.directory = directory
                self.status = status
                self.message = None

            def status_equals(self, other_status):
                return self.status == other_status

        # add mock data to the function calls that are essential
        stub_directory_status_valid = StubDirectoryStatus("valid", DirectoryStatus.NEW)
        stub_directory_status_invalid = StubDirectoryStatus("invalid", DirectoryStatus.INVALID)
        stub_directory_status_complete = StubDirectoryStatus("complete", DirectoryStatus.COMPLETE)
        stub_directory_status_partial = StubDirectoryStatus("partial", DirectoryStatus.PARTIAL)

        mock_parsing_handler.get_run_status_list.side_effect = [
            [stub_directory_status_valid,
             stub_directory_status_invalid,
             stub_directory_status_complete,
             stub_directory_status_partial]
        ]
        mock_parsing_handler.parse_and_validate.side_effect = ["Fake Sequencing Run"]

        # start
        upload.batch_upload_single_entry("fake_directory", force_upload=True)

        # assert calls are what we expect
        self.assertEqual(mock_validate_and_upload.call_count, 3, "Expected 3 calls to mock_validate_and_upload")
        expected_call_args = [
            call(stub_directory_status_valid, MODE_DEFAULT, False),
            call(stub_directory_status_complete, MODE_DEFAULT, False),
            call(stub_directory_status_partial, MODE_DEFAULT, False)
        ]
        self.assertEqual(mock_validate_and_upload.call_args_list, expected_call_args, "Call args do not match expected")



import unittest
from os import path


from iridauploader.core import uniform_file_count_validator
from iridauploader.parsers.miseq import parser as miseq_parser
from iridauploader.parsers.exceptions import SequenceFileError
from iridauploader import model

path_to_module = path.abspath(path.dirname(__file__))
if len(path_to_module) == 0:
    path_to_module = '.'


class TestValidateFileSizeMinimum(unittest.TestCase):
    """
    Testing the validate_uniform_file_count function
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @staticmethod
    def _make_seq_run_paired():
        """
        Make a sequencing run pointed at real data for the tests
        This dataset is a paired end run
        :return: SequencingRun object
        """
        files_1 = model.SequenceFile([
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "01-1111_S1_L001_R1_001.fastq.gz"),
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "01-1111_S1_L001_R2_001.fastq.gz"),
        ])
        files_2 = model.SequenceFile([
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "02-2222_S1_L001_R1_001.fastq.gz"),
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "02-2222_S1_L001_R2_001.fastq.gz"),
        ])
        files_3 = model.SequenceFile([
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "03-3333_S1_L001_R1_001.fastq.gz"),
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "03-3333_S1_L001_R2_001.fastq.gz"),
        ])
        sample_1 = model.Sample("test_sample", "description", 1)
        sample_1.sequence_file = files_1
        sample_2 = model.Sample("test_sample", "description", 1)
        sample_2.sequence_file = files_2
        sample_3 = model.Sample("test_sample", "description", 1)
        sample_3.sequence_file = files_3
        project = model.Project("test_project", [sample_1, sample_2, sample_3], "description")
        sequencing_run = model.SequencingRun({"layoutType": "PAIRED_END"}, [project], "miseq")
        return sequencing_run

    @staticmethod
    def _make_seq_run_single():
        """
        Make a sequencing run pointed at real data for the tests
        This dataset is a single end run
        :return: SequencingRun object
        """
        files_1 = model.SequenceFile([
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "01-1111_S1_L001_R1_001.fastq.gz"),
        ])
        files_2 = model.SequenceFile([
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "02-2222_S1_L001_R1_001.fastq.gz"),
        ])
        files_3 = model.SequenceFile([
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "03-3333_S1_L001_R1_001.fastq.gz"),
        ])
        sample_1 = model.Sample("test_sample", "description", 1)
        sample_1.sequence_file = files_1
        sample_2 = model.Sample("test_sample", "description", 1)
        sample_2.sequence_file = files_2
        sample_3 = model.Sample("test_sample", "description", 1)
        sample_3.sequence_file = files_3
        project = model.Project("test_project", [sample_1, sample_2, sample_3], "description")
        sequencing_run = model.SequencingRun({"layoutType": "SINGLE_END"}, [project], "miseq")
        return sequencing_run

    @staticmethod
    def _make_seq_run_mixed():
        """
        Make a sequencing run pointed at real data for the tests
        This dataset mixes single end and paired end runs, but identifies as paired end
        :return: SequencingRun object
        """
        files_1 = model.SequenceFile([
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "01-1111_S1_L001_R1_001.fastq.gz"),
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "01-1111_S1_L001_R2_001.fastq.gz"),
        ])
        files_2 = model.SequenceFile([
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "02-2222_S1_L001_R1_001.fastq.gz"),
        ])
        files_3 = model.SequenceFile([
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "03-3333_S1_L001_R1_001.fastq.gz"),
        ])
        sample_1 = model.Sample("test_sample", "description", 1)
        sample_1.sequence_file = files_1
        sample_2 = model.Sample("test_sample", "description", 1)
        sample_2.sequence_file = files_2
        sample_3 = model.Sample("test_sample", "description", 1)
        sample_3.sequence_file = files_3
        project = model.Project("test_project", [sample_1, sample_2, sample_3], "description")
        sequencing_run = model.SequencingRun({"layoutType": "PAIRED_END"}, [project], "miseq")
        return sequencing_run

    @staticmethod
    def _make_seq_run_paired_with_incorrect_file_name():
        """
        Make a sequencing run pointed at real data for the tests
        This dataset is a paired end run with a common user error of a misnamed file
        :return: SequencingRun object
        """
        files_1 = model.SequenceFile([
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "01-1111_S1_L001_R1_001.fastq.gz"),
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "01-1111_S1_L001_R2_001.fastq.gz"),
        ])
        files_2 = model.SequenceFile([
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "02-2222_S1_L001_R1_001.fastq.gz"),
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "02-2222x_S1_L001_R2_001.fastq.gz"),
        ])
        files_3 = model.SequenceFile([
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "03-3333_S1_L001_R1_001.fastq.gz"),
            path.join(path_to_module,
                      "fake_ngs_data", "Data", "Intensities", "BaseCalls", "03-3333_S1_L001_R2_001.fastq.gz"),
        ])
        sample_1 = model.Sample("test_sample", "description", 1)
        sample_1.sequence_file = files_1
        sample_2 = model.Sample("test_sample", "description", 1)
        sample_2.sequence_file = files_2
        sample_3 = model.Sample("test_sample", "description", 1)
        sample_3.sequence_file = files_3
        project = model.Project("test_project", [sample_1, sample_2, sample_3], "description")
        sequencing_run = model.SequencingRun({"layoutType": "PAIRED_END"}, [project], "miseq")
        return sequencing_run

    def test_run_valid_paired(self):
        sequencing_run = self._make_seq_run_paired()

        # Run code to test
        res = uniform_file_count_validator.validate_uniform_file_count(sequencing_run)

        # validate result
        self.assertTrue(res.is_valid(), "valid run is being detected as invalid")

    def test_run_valid_single(self):
        sequencing_run = self._make_seq_run_single()

        # Run code to test
        res = uniform_file_count_validator.validate_uniform_file_count(sequencing_run)

        # validate result
        self.assertTrue(res.is_valid(), "valid run is being detected as invalid")

    def test_run_valid_mixed_expecting_paired(self):
        sequencing_run = self._make_seq_run_mixed()

        # Run code to test
        res = uniform_file_count_validator.validate_uniform_file_count(sequencing_run)

        # validate result
        self.assertFalse(res.is_valid(), "invalid run is being detected as valid")

    def test_run_valid_mixed_expecting_single(self):
        sequencing_run = self._make_seq_run_mixed()
        # change metadata to expect single end run
        sequencing_run.metadata["layoutType"] = "SINGLE_END"

        # Run code to test
        res = uniform_file_count_validator.validate_uniform_file_count(sequencing_run)

        # validate result
        self.assertFalse(res.is_valid(), "invalid run is being detected as valid")

    def test_run_invalid_name_error(self):
        # Build sequencing run from real data with common user error
        # One of the files has an extra character in the filename in a spot that is within spec to be parsed,
        # but results in mixing single end and paired end files in the same run.
        sample_sheet_path = path.join(
            path_to_module, "fake_ngs_data_name_error", "SampleSheet.csv")
        parser_instance = miseq_parser.Parser()
        sequencing_run = parser_instance.get_sequencing_run(sample_sheet_path)

        # Run code to test
        res = uniform_file_count_validator.validate_uniform_file_count(sequencing_run)

        # validate result
        self.assertFalse(res.is_valid(), "invalid run is being detected as valid")
        self.assertEqual(
            str(res.error_list),
            str([SequenceFileError(
                'File count for sample `02-2222` does not match the expected file count `2`. Please verify your data.'
            )]))



import unittest
from unittest.mock import patch, MagicMock
from os import path

from iridauploader.api import UPLOAD_MODES, MODE_DEFAULT, MODE_FAST5, MODE_ASSEMBLIES
from iridauploader.core import upload_helpers
from iridauploader.model import DirectoryStatus, SequencingRun, Project, Sample
from iridauploader import progress
from iridauploader import parsers
from iridauploader.api.exceptions import FileError, IridaResourceError, IridaConnectionError

path_to_module = path.abspath(path.dirname(__file__))
if len(path_to_module) == 0:
    path_to_module = '.'


class TestSetAndWriteDirectoryStatus(unittest.TestCase):
    """
    Tests core.upload_helpers._set_and_write_directory_status
    """

    class StubDirectoryStatus:
        status = None
        message = ""
        run_id = None

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @patch("iridauploader.core.upload_helpers.progress")
    def test_valid_write(self, mock_progress):
        """
        Tessts a valid write attempt
        :param mock_progress:
        :return:
        """
        mock_status = "status"
        mock_message = "message"
        stub_dir_status = self.StubDirectoryStatus()
        # mock main call to test
        mock_progress.write_directory_status.side_effect = [None]
        # run function
        upload_helpers._set_and_write_directory_status(stub_dir_status, mock_status, mock_message)
        # verify write
        mock_progress.write_directory_status.assert_called_with(stub_dir_status)

    @patch("iridauploader.core.upload_helpers.progress")
    def test_valid_write_error(self, mock_progress):
        """
        Tessts a valid write attempt of ERROR, where the upload has not begun
        :param mock_progress:
        :return:
        """
        mock_initial_status = DirectoryStatus.ERROR
        mock_expected_status = DirectoryStatus.ERROR
        mock_message = "message"
        stub_dir_status = self.StubDirectoryStatus()
        stub_dir_status.run_id = None  # Upload has not started
        # mock main call to test
        mock_progress.write_directory_status.side_effect = [None]
        # run function
        upload_helpers._set_and_write_directory_status(stub_dir_status, mock_initial_status, mock_message)
        # verify write
        stub_dir_status.status = mock_expected_status
        mock_progress.write_directory_status.assert_called_with(stub_dir_status)

    @patch("iridauploader.core.upload_helpers.progress")
    def test_valid_write_partial(self, mock_progress):
        """
        Tessts a valid write attempt of PARTIAL, ERROR is given but the run has started
        :param mock_progress:
        :return:
        """
        mock_initial_status = DirectoryStatus.ERROR
        mock_expected_status = DirectoryStatus.ERROR
        mock_message = "message"
        stub_dir_status = self.StubDirectoryStatus()
        stub_dir_status.run_id = 123  # Upload has not started
        # mock main call to test
        mock_progress.write_directory_status.side_effect = [None]
        # run function
        upload_helpers._set_and_write_directory_status(stub_dir_status, mock_initial_status, mock_message)
        # verify write
        stub_dir_status.status = mock_expected_status
        mock_progress.write_directory_status.assert_called_with(stub_dir_status)

    @patch("iridauploader.core.upload_helpers.progress.write_directory_status")
    def test_invalid_write(self, mock_progress):
        """
        Tessts a valid write attempt
        :param mock_progress:
        :return:
        """
        mock_status = "status"
        mock_message = "message"
        stub_dir_status = self.StubDirectoryStatus()
        # mock main call to test
        mock_progress.side_effect = progress.exceptions.DirectoryError("", "")
        # run function
        with self.assertRaises(progress.exceptions.DirectoryError):
            upload_helpers._set_and_write_directory_status(stub_dir_status, mock_status, mock_message)


class TestDirectoryHasReadonlyConflict(unittest.TestCase):
    """
    Tests core.upload_helpers.directory_has_readonly_conflict
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @patch("os.access")
    @patch("iridauploader.core.upload_helpers.config")
    def test_case_matrix(self, mock_config, mock_access):
        """
        Tests all possible cases for readonly config and access response
        :param mock_config:
        :param mock_access:
        :return:
        """
        # Matrix of tests
        # config:  T  F
        # access
        # T        X  X
        # F        X  O
        mock_config.read_config_option.side_effect = [
            True, True, False, False
        ]
        mock_access.side_effect = [
            True, False, True, False
        ]

        self.assertEqual(False, upload_helpers.directory_has_readonly_conflict(""))
        self.assertEqual(False, upload_helpers.directory_has_readonly_conflict(""))
        self.assertEqual(False, upload_helpers.directory_has_readonly_conflict(""))
        self.assertEqual(True, upload_helpers.directory_has_readonly_conflict(""))


class TestParseAndValidate(unittest.TestCase):
    """
    Tests core.upload_helpers.parse_and_validate
    """

    class StubDirectoryStatus:
        directory = "dir"

        def get_sample_status_list(self):
            return ['item']

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @patch("iridauploader.core.upload_helpers._set_and_write_directory_status")
    @patch("iridauploader.core.upload_helpers.parsing_handler")
    def test_valid_parse(self, mock_parsing_handler, mock_set_and_write):
        """
        verifies parse and validate was called,
        :param mock_parsing_handler:
        :param mock_set_and_write:
        :return:
        """
        stub_directory_status = self.StubDirectoryStatus()
        mock_parsing_handler.parse_and_validate.side_effect = ["return_value"]
        mock_set_and_write.side_effect = [True]

        self.assertEqual("return_value", upload_helpers.parse_and_validate(
            directory_status=stub_directory_status, parse_as_partial=False))

        mock_parsing_handler.parse_and_validate.assert_called_with(stub_directory_status.directory)

    @patch("iridauploader.core.upload_helpers.set_uploaded_samples_to_skip")
    @patch("iridauploader.core.upload_helpers._set_and_write_directory_status")
    @patch("iridauploader.core.upload_helpers.parsing_handler")
    def test_valid_partial_parse(self, mock_parsing_handler, mock_set_and_write, mock_set_uploaded_samples_to_skip):
        """
        verifies parse and validate was called with partial run handling
        :param mock_parsing_handler:
        :param mock_set_and_write:
        :return:
        """
        stub_directory_status = self.StubDirectoryStatus()
        mock_parsing_handler.parse_and_validate.side_effect = ["return_value"]
        mock_set_and_write.side_effect = [True]
        mock_set_uploaded_samples_to_skip.side_effect = ["modified_return_value"]

        self.assertEqual("modified_return_value", upload_helpers.parse_and_validate(
            directory_status=stub_directory_status, parse_as_partial=True))

        mock_parsing_handler.parse_and_validate.assert_called_with(stub_directory_status.directory)
        mock_set_uploaded_samples_to_skip.assert_called_with("return_value",
                                                             stub_directory_status.get_sample_status_list())

    @patch("iridauploader.core.upload_helpers._set_and_write_directory_status")
    @patch("iridauploader.core.upload_helpers.parsing_handler")
    def test_invalid_directory(self, mock_parsing_handler, mock_set_and_write):
        """
        verifies parse and validate was called, throwing DirectoryError
        and _set_and_write_directory_status is called with ERROR
        :param mock_parsing_handler:
        :param mock_set_and_write:
        :return:
        """
        stub_directory_status = self.StubDirectoryStatus()
        mock_parsing_handler.parse_and_validate.side_effect = [parsers.exceptions.DirectoryError("", "")]
        mock_set_and_write.side_effect = [True, True]

        with self.assertRaises(parsers.exceptions.DirectoryError):
            upload_helpers.parse_and_validate(directory_status=stub_directory_status, parse_as_partial=False)

        mock_parsing_handler.parse_and_validate.assert_called_with(stub_directory_status.directory)
        mock_set_and_write.assert_called_with(stub_directory_status, DirectoryStatus.ERROR,
                                              "ERROR! An error occurred with directory '', with message: ")

    @patch("iridauploader.core.upload_helpers._set_and_write_directory_status")
    @patch("iridauploader.core.upload_helpers.parsing_handler")
    def test_invalid_validation(self, mock_parsing_handler, mock_set_and_write):
        """
        verifies parse and validate was called, throwing ValidationError
        and _set_and_write_directory_status is called with ERROR
        :param mock_parsing_handler:
        :param mock_set_and_write:
        :return:
        """

        class StubValidationResult:
            error_list = []

        stub_validation_result = StubValidationResult()
        stub_directory_status = self.StubDirectoryStatus()
        mock_parsing_handler.parse_and_validate.side_effect = [parsers.exceptions.ValidationError(
            "",
            stub_validation_result)]
        mock_set_and_write.side_effect = [True, True]

        with self.assertRaises(parsers.exceptions.ValidationError):
            upload_helpers.parse_and_validate(directory_status=stub_directory_status, parse_as_partial=False)

        mock_parsing_handler.parse_and_validate.assert_called_with(stub_directory_status.directory)
        mock_set_and_write.assert_called_with(stub_directory_status, DirectoryStatus.ERROR,
                                              'ERROR! Errors occurred during validation with message: , Error list: []')


class TestSetUploadedSamplesToSkip(unittest.TestCase):
    """
    Tests core.upload_helpers.set_uploaded_samples_to_skip
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_remove_one(self):
        """
        Test setting a single sample to skip in a project

        :return:
        """
        seq_run = SequencingRun(None, sequencing_run_type="test",
                                project_list=[
                                    Project(sample_list=[
                                        Sample("one"),
                                        Sample("two"),
                                        Sample("three")
                                    ], id=1)
                                ])
        sample_status_list = [DirectoryStatus.SampleStatus(sample_name="one", project_id="1", uploaded=False),
                              DirectoryStatus.SampleStatus(sample_name="two", project_id="1", uploaded=True),
                              DirectoryStatus.SampleStatus(sample_name="three", project_id="1", uploaded=False)]

        res = upload_helpers.set_uploaded_samples_to_skip(seq_run, sample_status_list)

        res_samples = res.project_list[0].sample_list
        self.assertEqual(res_samples[0].skip, False)
        self.assertEqual(res_samples[1].skip, True)
        self.assertEqual(res_samples[2].skip, False)


class TestVerifyUploadMode(unittest.TestCase):
    """
    Tests core.upload_helpers.verify_upload_mode
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_valid_upload_modes(self):
        """
        No exceptions being thrown means it is working as intended
        :return:
        """
        upload_helpers.verify_upload_mode(MODE_DEFAULT)
        upload_helpers.verify_upload_mode(MODE_FAST5)
        upload_helpers.verify_upload_mode(MODE_ASSEMBLIES)

    def test_invalid_upload_mode(self):
        """
        verify exception is thrown
        :return:
        """
        # verify mode doesn't exist

        mode = "invalid_mode"
        self.assertFalse(mode in UPLOAD_MODES)

        with self.assertRaises(Exception):
            upload_helpers.verify_upload_mode(mode)


class TestInitFileStatusListFromSequencingRun(unittest.TestCase):
    """
    Tests core.upload_helpers.init_file_status_list_from_sequencing_run
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @patch("iridauploader.core.upload_helpers._set_and_write_directory_status")
    def test_all_functions_called(self, mock_set_and_write):
        """
        verifies all functions are called, and _set_and_write_directory_status is DirectoryStatus.PARTIAL
        :param mock_set_and_write:
        :return:
        """
        mock_dir_status = MagicMock()
        mock_dir_status.init_file_status_list_from_sequencing_run.side_effect = [True]
        mock_set_and_write.side_effect = [True]

        upload_helpers.init_file_status_list_from_sequencing_run("seqrun", mock_dir_status)

        mock_dir_status.init_file_status_list_from_sequencing_run.assert_called_with("seqrun")


class TestInitializeApi(unittest.TestCase):
    """
    Tests core.upload_helpers.initialize_api
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @patch("iridauploader.core.upload_helpers._set_and_write_directory_status")
    @patch("iridauploader.core.upload_helpers.api_handler")
    def test_valid_init(self, mock_api_handler, mock_set_and_write):
        """
        verifies init was called, and _set_and_write_directory_status was not called
        :param mock_api_handler:
        :param mock_set_and_write:
        :return:
        """
        mock_api_handler.initialize_api_from_config.side_effect = [True]
        mock_set_and_write.side_effect = [True]

        upload_helpers.initialize_api(directory_status='status')

        mock_api_handler.initialize_api_from_config.assert_called_with()
        mock_set_and_write.assert_not_called()

    @patch("iridauploader.core.upload_helpers._set_and_write_directory_status")
    @patch("iridauploader.core.upload_helpers.api_handler")
    def test_invalid_init(self, mock_api_handler, mock_set_and_write):
        """
        verifies init was called, and because of error, _set_and_write_directory_status is DirectoryStatus.ERROR
        :param mock_api_handler:
        :param mock_set_and_write:
        :return:
        """
        mock_api_handler.initialize_api_from_config.side_effect = [IridaConnectionError]
        mock_set_and_write.side_effect = [True]

        with self.assertRaises(IridaConnectionError):
            upload_helpers.initialize_api(directory_status='status')

        mock_api_handler.initialize_api_from_config.assert_called_with()

        mock_set_and_write.assert_called_with("status", DirectoryStatus.ERROR,
                                              'ERROR! Could not initialize irida api. Errors: ()')


class TestIridaPrepAndValidation(unittest.TestCase):
    """
    Tests core.upload_helpers.irida_prep_and_validation
    """

    class StubValidationResult:
        valid = True
        error_list = []

        def is_valid(self):
            return self.valid

        @staticmethod
        def error_count():
            return 0

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @patch("iridauploader.core.upload_helpers._set_and_write_directory_status")
    @patch("iridauploader.core.upload_helpers.api_handler")
    def test_valid_prep_and_validation(self, mock_api_handler, mock_set_and_write):
        """
        verifies prep_and_validation was called, and _set_and_write_directory_status was not called
        :param mock_api_handler:
        :param mock_set_and_write:
        :return:
        """
        stub_validation_result = self.StubValidationResult()
        mock_api_handler.prepare_and_validate_for_upload.side_effect = [stub_validation_result]
        mock_set_and_write.side_effect = [True]

        upload_helpers.irida_prep_and_validation("seqrun", "")

        mock_api_handler.prepare_and_validate_for_upload.assert_called_with("seqrun")
        mock_set_and_write.assert_not_called()

    @patch("iridauploader.core.upload_helpers._set_and_write_directory_status")
    @patch("iridauploader.core.upload_helpers.api_handler")
    def test_invalid_connection(self, mock_api_handler, mock_set_and_write):
        """
        verifies prep_and_validation was called,
        and _set_and_write_directory_status was called with ERROR
        :param mock_api_handler:
        :param mock_set_and_write:
        :return:
        """
        mock_api_handler.prepare_and_validate_for_upload.side_effect = [IridaConnectionError]
        mock_set_and_write.side_effect = [True]

        with self.assertRaises(IridaConnectionError):
            upload_helpers.irida_prep_and_validation("seqrun", "")

        mock_api_handler.prepare_and_validate_for_upload.assert_called_with("seqrun")
        mock_set_and_write.assert_called_with("", DirectoryStatus.ERROR, 'Lost connection to Irida. Errors: ()')

    @patch("iridauploader.core.upload_helpers._set_and_write_directory_status")
    @patch("iridauploader.core.upload_helpers.api_handler")
    def test_invalid_validation(self, mock_api_handler, mock_set_and_write):
        """
        verifies prep_and_validation was called,
        and _set_and_write_directory_status was called with ERROR
        :param mock_api_handler:
        :param mock_set_and_write:
        :return:
        """
        stub_validation_result = self.StubValidationResult()
        stub_validation_result.valid = False
        mock_api_handler.prepare_and_validate_for_upload.side_effect = [stub_validation_result]
        mock_set_and_write.side_effect = [True]

        with self.assertRaises(Exception):
            upload_helpers.irida_prep_and_validation("seqrun", "")

        mock_api_handler.prepare_and_validate_for_upload.assert_called_with("seqrun")
        mock_set_and_write.assert_called_with("", DirectoryStatus.ERROR,
                                              'Sequencing run can not be uploaded, Errors: []')


class TestUploadSequencingRun(unittest.TestCase):
    """
    Tests core.upload_helpers.upload_sequencing_run
    """

    class StubDirectoryStatus:
        status = None
        message = ""
        run_id = None

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @patch("iridauploader.core.upload_helpers._set_and_write_directory_status")
    @patch("iridauploader.core.upload_helpers.api_handler")
    def test_valid_upload(self, mock_api_handler, mock_set_and_write):
        """
        verifies upload was called, and _set_and_write_directory_status is DirectoryStatus.COMPLETE
        :param mock_api_handler:
        :param mock_set_and_write:
        :return:
        """
        mock_api_handler.upload_sequencing_run.side_effect = [True]
        mock_set_and_write.side_effect = [True]

        upload_helpers.upload_sequencing_run(directory_status='status',
                                             sequencing_run='run',
                                             upload_mode='mode')

        mock_api_handler.upload_sequencing_run.assert_called_with(directory_status='status',
                                                                  sequencing_run='run',
                                                                  upload_mode='mode',
                                                                  run_id=None)
        mock_set_and_write.assert_called_with("status", DirectoryStatus.COMPLETE)

    @patch("iridauploader.core.upload_helpers._set_and_write_directory_status")
    @patch("iridauploader.core.upload_helpers.api_handler")
    def test_valid_partial_upload(self, mock_api_handler, mock_set_and_write):
        """
        verifies upload was called, and _set_and_write_directory_status is DirectoryStatus.COMPLETE
        :param mock_api_handler:
        :param mock_set_and_write:
        :return:
        """

        class MockDirStatus:
            run_id = 1

        mock_api_handler.upload_sequencing_run.side_effect = [True]
        mock_set_and_write.side_effect = [True]
        mock_directory_status = MockDirStatus()

        upload_helpers.upload_sequencing_run(directory_status=mock_directory_status,
                                             sequencing_run='run',
                                             upload_mode='mode',
                                             upload_from_partial=True)

        mock_api_handler.upload_sequencing_run.assert_called_with(directory_status=mock_directory_status,
                                                                  sequencing_run='run',
                                                                  upload_mode='mode',
                                                                  run_id=1)
        mock_set_and_write.assert_called_with(mock_directory_status, DirectoryStatus.COMPLETE)

    @patch("iridauploader.core.upload_helpers._set_and_write_directory_status")
    @patch("iridauploader.core.upload_helpers.api_handler")
    def test_invalid_connection(self, mock_api_handler, mock_set_and_write):
        """
        tests catching and raising IridaConnectionError
        and _set_and_write_directory_status is DirectoryStatus.ERROR
        :param mock_api_handler:
        :param mock_set_and_write:
        :return:
        """
        mock_api_handler.upload_sequencing_run.side_effect = [IridaConnectionError]
        mock_set_and_write.side_effect = [True]

        # Init directory status as a new run
        stub_directory_status = self.StubDirectoryStatus()
        stub_directory_status.status = DirectoryStatus.NEW
        stub_directory_status.run_id = None

        with self.assertRaises(IridaConnectionError):
            upload_helpers.upload_sequencing_run(directory_status=stub_directory_status,
                                                 sequencing_run='run',
                                                 upload_mode='mode')

        mock_api_handler.upload_sequencing_run.assert_called_with(directory_status=stub_directory_status,
                                                                  sequencing_run='run',
                                                                  upload_mode='mode',
                                                                  run_id=None)
        mock_set_and_write.assert_called_with(stub_directory_status,
                                              DirectoryStatus.ERROR,
                                              'Lost connection to Irida. Errors: ()')

    @patch("iridauploader.core.upload_helpers._set_and_write_directory_status")
    @patch("iridauploader.core.upload_helpers.api_handler")
    def test_invalid_resource(self, mock_api_handler, mock_set_and_write):
        """
        tests catching and raising IridaResourceError
        and _set_and_write_directory_status is DirectoryStatus.COMPLETE
        :param mock_api_handler:
        :param mock_set_and_write:
        :return:
        """
        mock_api_handler.upload_sequencing_run.side_effect = [IridaResourceError("")]
        mock_set_and_write.side_effect = [True]

        with self.assertRaises(IridaResourceError):
            upload_helpers.upload_sequencing_run(directory_status='status',
                                                 sequencing_run='run',
                                                 upload_mode='mode')

        mock_api_handler.upload_sequencing_run.assert_called_with(directory_status='status',
                                                                  sequencing_run='run',
                                                                  upload_mode='mode',
                                                                  run_id=None)
        mock_set_and_write.assert_called_with("status", DirectoryStatus.ERROR,
                                              "Could not access IRIDA resource Errors: ('',)")

    @patch("iridauploader.core.upload_helpers._set_and_write_directory_status")
    @patch("iridauploader.core.upload_helpers.api_handler")
    def test_invalid_file(self, mock_api_handler, mock_set_and_write):
        """
        tests catching and raising FileError
        and _set_and_write_directory_status is DirectoryStatus.COMPLETE
        :param mock_api_handler:
        :param mock_set_and_write:
        :return:
        """
        mock_api_handler.upload_sequencing_run.side_effect = [FileError]
        mock_set_and_write.side_effect = [True]

        with self.assertRaises(FileError):
            upload_helpers.upload_sequencing_run(directory_status='status',
                                                 sequencing_run='run',
                                                 upload_mode='mode')

        mock_api_handler.upload_sequencing_run.assert_called_with(directory_status='status',
                                                                  sequencing_run='run',
                                                                  upload_mode='mode',
                                                                  run_id=None)
        mock_set_and_write.assert_called_with("status", DirectoryStatus.ERROR,
                                              'Could not upload file to IRIDA. Errors: ()')


class TestSetRunDelayed(unittest.TestCase):
    """
    Tests core.upload_helpers.set_run_delayed
    """

    class StubDirectoryStatus:
        directory = "dir"

        def get_sample_status_list(self):
            return ['item']

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @patch("iridauploader.core.upload_helpers._set_and_write_directory_status")
    def test_valid(self, mock_set_and_write):
        """
        verifies parse and validate was called,
        :param mock_set_and_write:
        :return:
        """
        stub_directory_status = self.StubDirectoryStatus()
        # mock_parsing_handler.parse_and_validate.side_effect = ["return_value"]
        mock_set_and_write.side_effect = [True]

        upload_helpers.set_run_delayed(directory_status=stub_directory_status)

        mock_set_and_write.assert_called_with(stub_directory_status, DirectoryStatus.DELAYED)



from os import path
from collections import OrderedDict
from copy import deepcopy
import logging

import iridauploader.model as model
from iridauploader.parsers import exceptions
from iridauploader.parsers import common


def parse_metadata(sample_list):
    """
    Determine if samples are paired or single end, and return metadata to match

    :param sample_list: List of Sample objects
    :return: metadata dictionary
    """
    # add the layout type to the sequencing run so we know if it is paired or single end
    if sample_list[0].sequence_file.is_paired_end():
        metadata = {'layoutType': 'PAIRED_END'}
    else:
        metadata = {'layoutType': 'SINGLE_END'}

    return metadata


def verify_sample_sheet_file_names_in_file_list(sample_sheet_file, run_data_directory_file_list):
    """
    Given a sample sheet, and a list of files in a directory,
    verify that all the files on the sheet exist in the file list

    If a file is missing, a SampleSheetError is raised

    :param sample_sheet_file:
    :param run_data_directory_file_list:
    :return:
    """
    sample_list = _parse_samples(sample_sheet_file)

    for sample in sample_list:
        sample_dict = sample.get_uploadable_dict()
        paired_end_read = len(sample_dict['File_Reverse']) > 0

        # Check if file names are in the files we found in the directory file list
        if sample_dict['File_Forward'] not in run_data_directory_file_list:
            raise exceptions.SampleSheetError(
                ("Your sample sheet is malformed. {} Does not match any file list in sample sheet file"
                 "".format(sample_dict['File_Forward'])),
                sample_sheet_file
            )
        if paired_end_read and sample_dict['File_Reverse'] not in run_data_directory_file_list:
            raise exceptions.SampleSheetError(
                ("Your sample sheet is malformed. {} Does not match any file list in sample sheet file"
                 "".format(sample_dict['File_Reverse'])),
                sample_sheet_file
            )


def build_sample_list_from_sample_sheet_with_abs_path(sample_sheet_file):
    """
    Create a list of Sample objects, where each SequenceFile object has an absolute file path

    :param sample_sheet_file:
    :return:
    """
    sample_list = _parse_samples(sample_sheet_file)
    # Data directory is used if file names on sample sheet are not absolute paths (in directory files)
    data_dir = path.dirname(sample_sheet_file)
    sample_sheet_dir_file_list = common.get_file_list(data_dir)

    for sample in sample_list:
        sample_dict = sample.get_uploadable_dict()
        paired_end_read = len(sample_dict['File_Reverse']) > 0

        # create file list of full paths
        file_list = []
        # If file is not an abspath already, make it an abspath from filename + data dir
        if path.isabs(sample_dict['File_Forward']):
            file_list.append(sample_dict['File_Forward'])
        elif sample_dict['File_Forward'] in sample_sheet_dir_file_list:
            sample_dict['File_Forward'] = path.join(path.abspath(data_dir), sample_dict['File_Forward'])

            file_list.append(sample_dict['File_Forward'])
        else:
            raise exceptions.SampleSheetError(
                ("Your sample sheet is malformed. {} Does not match any file in the directory {}"
                 "".format(sample_dict['File_Forward'], data_dir)),
                sample_sheet_file)

        # reverse file is same as for forward file
        if paired_end_read:
            if path.isabs(sample_dict['File_Reverse']):
                file_list.append(sample_dict['File_Reverse'])
            elif sample_dict['File_Reverse'] in sample_sheet_dir_file_list:
                sample_dict['File_Reverse'] = path.join(path.abspath(data_dir), sample_dict['File_Reverse'])
                file_list.append(sample_dict['File_Reverse'])
            else:
                raise exceptions.SampleSheetError(
                    ("Your sample sheet is malformed. {} Does not match any file in the directory {}"
                     "".format(sample_dict['File_Reverse'], data_dir)),
                    sample_sheet_file)

        # Create sequence file object and attach to sample
        sq = model.SequenceFile(file_list=file_list)
        sample.sequence_file = deepcopy(sq)

    return sample_list


def build_sample_list_from_sample_sheet_no_verify(sample_sheet_file):
    """
    Create a list of Sample objects, file existence is not verified before SequenceFile is created
    this is used when a pre-generated file list is used (e.g. cloud deployment)

    :param sample_sheet_file:
    :return:
    """
    sample_list = _parse_samples(sample_sheet_file)

    for sample in sample_list:

        sample_dict = sample.get_uploadable_dict()
        # create file list
        file_list = [sample_dict['File_Forward']]

        # if paired end add file to file list
        paired_end_read = len(sample_dict['File_Reverse']) > 0
        if paired_end_read:
            file_list.append(sample_dict['File_Reverse'])

        # Create sequence file object and attach to sample
        sq = model.SequenceFile(file_list=file_list)
        sample.sequence_file = deepcopy(sq)

    return sample_list


def only_single_or_paired_in_sample_list(sample_list):
    """
    Given a list of Sample objects, verifies there are only one type (single end or paired end)

    :param sample_list:
    :return: Boolean
    """
    has_single = False
    has_paired = False

    for sample in sample_list:
        if sample.sequence_file.is_paired_end():
            has_paired = True
        else:
            has_single = True

    return not (has_single and has_paired)


def _parse_samples(sample_sheet_file):

    """
    Parse all the lines under "[Data]" in .csv file

    arguments:
            sample_sheet_file -- path to SampleSheet.csv

    returns	a list containing Sample objects that have been created by a
        dictionary from the parsed out key:pair values from .csv file
    """

    logging.info("Reading data from sample sheet {}".format(sample_sheet_file))

    csv_reader = common.get_csv_reader(sample_sheet_file)
    # start with an ordered dictionary so that keys are ordered in the same
    # way that they are inserted.
    sample_dict = OrderedDict()
    sample_list = []

    sample_key_list = ['Sample_Name', 'Project_ID', 'File_Forward', 'File_Reverse']

    # initialize dictionary keys from first line (data headers/attributes)
    set_attributes = False
    for line in csv_reader:

        if set_attributes:
            for item in line:

                if item in sample_key_list:
                    key_name = item
                    sample_dict[key_name] = ""

            break

        if "[Data]" in line:
            set_attributes = True

    # fill in values for keys. line is currently below the [Data] headers
    for sample_number, line in enumerate(csv_reader):
        # if the line is empty (like a blank line at the end of the file) continue
        if not line:
            continue

        if len(sample_dict.keys()) != len(line):
            """
            if there is one more Data header compared to the length of
            data values then add an empty string to the end of data values
            i.e the File_Reverse will be empty string
            assumes the last Data header is going to be the File_Reverse
            this handles the case where the last trailing comma is trimmed when
            doing a single end run
            """
            if len(sample_dict.keys()) - len(line) == 1:
                line.append("")
            else:
                raise exceptions.SampleSheetError(
                    ("Your sample sheet is malformed. Expected to find {} "
                     "columns in the [Data] section, but only found {} columns "
                     "for line {}.".format(len(sample_dict.keys()), len(line), line)),
                    sample_sheet_file
                )

        for index, key in enumerate(sample_dict.keys()):
            value = line[index].strip()

            # Keys other than 'File_Reverse' cannot be empty
            if len(value) == 0:  # no value
                if key != 'File_Reverse':
                    raise exceptions.SampleSheetError(
                        ("Your sample sheet is malformed. {} in the [Data] section cannot be empty."
                         "".format(key)),
                        sample_sheet_file
                    )

            sample_dict[key] = value

        new_sample_dict = deepcopy(sample_dict)
        new_sample_name = new_sample_dict['Sample_Name']
        new_sample_project = new_sample_dict['Project_ID']
        new_sample_dict['sample_project'] = new_sample_project
        del new_sample_dict['Sample_Name']
        del new_sample_dict['Project_ID']

        sample = model.Sample(
            sample_name=new_sample_name,
            description="",
            sample_number=sample_number + 1,
            samp_dict=new_sample_dict)

        sample_list.append(sample)

    return sample_list



class ValidationResult:

    def __init__(self):
        self._error_list = []

    def add_error(self, error):
        self._error_list.append(error)

    def error_count(self):
        return len(self._error_list)

    @property
    def error_list(self):
        return self._error_list

    def is_valid(self):
        return self.error_count() == 0



import unittest
from unittest.mock import patch
from os import path
from csv import reader
from io import StringIO

from iridauploader import parsers
import iridauploader.parsers.nextseq.sample_parser as sample_parser
from iridauploader.parsers.exceptions import SampleSheetError, SequenceFileError
import iridauploader.model as model

path_to_module = path.abspath(path.dirname(__file__))
if len(path_to_module) == 0:
    path_to_module = '.'


class TestParseMetadata(unittest.TestCase):
    """
    Test building meta data from sample sheets with a csv reader
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @patch("iridauploader.parsers.common.get_csv_reader")
    def test_parse_metadata_paired_valid(self, mock_csv_reader):
        """
        When given a valid directory, ensure valid metadata is built
        paired end reads
        :return:
        """
        h_field_values = (
            "IEMFileVersion,4\n"
            "Investigator Name,Test Name\n"
            "Experiment Name,Some_Test_Data\n"
            "Date,2015-05-14\n"
            "Workflow,GenerateFASTQ\n"
            "Application,FASTQ Only\n"
            "Assay,ASDF\n"
            "Description,12-34\n"
            "Chemistry,Yes\n"
        )

        reads = (
            "251\n"
            "251\n"
        )

        d_headers = ("Sample_ID,Sample_Name,Sample_Plate,Sample_Well,"
                     "I7_Index_ID,index,I5_Index_ID,index2,Sample_Project,"
                     "Description")

        d_field_values = (
            "15-0318,,2015-08-05-SE,A01,N701,TAAGGCGA,S502,CTCTCTAT,203\n"
            "15-0455,,2015-08-05-SE,B01,N701,TAAGGCGA,S503,TATCCTCT,203\n"
            "15-0462,,2015-08-05-SE,C01,N701,TAAGGCGA,S505,GTAAGGAG,203\n"
        )

        file_contents_str = (
            "[Header]\n"
            "{h_field_values}\n"
            "[Reads]\n"
            "{reads}\n"
            "[Data]\n"
            "{d_headers}\n"
            "{d_field_values}"
        ).format(h_field_values=h_field_values,
                 reads=reads,
                 d_headers=d_headers,
                 d_field_values=d_field_values)

        # converts string as a pseudo file / memory file
        sample_sheet_file = StringIO(file_contents_str)

        # the call to get_csv_reader() inside parse_samples() will return
        # items inside side_effect
        mock_csv_reader.side_effect = [reader(sample_sheet_file)]

        metadata = sample_parser.parse_metadata(None)
        # The meta data we care about the most
        self.assertEqual(metadata['readLengths'], "251")
        self.assertEqual(metadata['layoutType'], "PAIRED_END")
        # Other meta data should also be here
        self.assertEqual(metadata['iemfileversion'], "4")
        self.assertEqual(metadata['investigatorName'], "Test Name")
        self.assertEqual(metadata['experimentName'], "Some_Test_Data")
        self.assertEqual(metadata['date'], "2015-05-14")
        self.assertEqual(metadata['workflow'], "GenerateFASTQ")
        self.assertEqual(metadata['application'], "FASTQ Only")
        self.assertEqual(metadata['assay'], "ASDF")
        self.assertEqual(metadata['description'], "12-34")
        self.assertEqual(metadata['chemistry'], "Yes")

    @patch("iridauploader.parsers.common.get_csv_reader")
    def test_parse_metadata_single_valid(self, mock_csv_reader):
        """
        When given a valid directory, ensure valid metadata is built
        single end reads
        :return:
        """
        h_field_values = (
            "IEMFileVersion, 4\n"
            "Investigator Name, Test Name\n"
            "Experiment Name, Some_Test_Data\n"
            "Date, 2015-05-14\n"
            "Workflow, GenerateFASTQ\n"
            "Application, FASTQ Only\n"
            "Assay, ASDF\n"
            "Description, 12-34\n"
            "Chemistry, Yes\n"
        )

        reads = (
            "251\n"
        )

        d_headers = ("Sample_ID,Sample_Name,Sample_Plate,Sample_Well,"
                     "I7_Index_ID,index,I5_Index_ID,index2,Sample_Project,"
                     "Description")

        d_field_values = (
            "15-0318,,2015-08-05-SE,A01,N701,TAAGGCGA,S502,CTCTCTAT,203\n"
            "15-0455,,2015-08-05-SE,B01,N701,TAAGGCGA,S503,TATCCTCT,203\n"
            "15-0462,,2015-08-05-SE,C01,N701,TAAGGCGA,S505,GTAAGGAG,203\n"
        )

        file_contents_str = (
            "[Header]\n"
            "{h_field_values}\n"
            "[Reads]\n"
            "{reads}\n"
            "[Data]\n"
            "{d_headers}\n"
            "{d_field_values}"
        ).format(h_field_values=h_field_values,
                 reads=reads,
                 d_headers=d_headers,
                 d_field_values=d_field_values)

        # converts string as a pseudo file / memory file
        sample_sheet_file = StringIO(file_contents_str)

        # the call to get_csv_reader() inside parse_samples() will return
        # items inside side_effect
        mock_csv_reader.side_effect = [reader(sample_sheet_file)]

        metadata = sample_parser.parse_metadata(None)
        self.assertEqual(metadata['layoutType'], "SINGLE_END")

    def test_parse_metadata(self):
        """
        Testing the parsing meta data with actual files, instead of mocked files
        :return:
        """
        sheet_file = path.join(path_to_module, "fake_nextseq_run",
                               "SampleSheetClassic.csv")
        meta_data = sample_parser.parse_metadata(sheet_file)

        correct_metadata = {"readLengths": "151",
                            "assay": "Nextera XT",
                            "description": "Contact Experiment",
                            "application": "NextSeq FASTQ Only",
                            "investigatorName": "Ikari Yui",
                            "adapter": "CTGTCTCTTATACACATCT",
                            "workflow": "GenerateFASTQ",
                            "iemfileversion": "5",
                            "date": "2000-09-13",
                            "experimentName": "NGS-Adam",
                            "chemistry": "Amplicon",
                            "layoutType": "PAIRED_END"}

        self.assertEqual(correct_metadata, meta_data)

    def test_parse_metadata_extra_commas(self):
        """
        Tests parsing meta data when extra commas are in the file
        Testing the parsing meta data with actual files, instead of mocked files
        :return:
        """
        sheet_file = path.join(path_to_module, "testValidSheetTrailingCommas",
                               "SampleSheetClassic.csv")
        meta_data = sample_parser.parse_metadata(sheet_file)

        correct_metadata = {"readLengths": "301",
                            "assay": "TruSeq HT",
                            "description": "252",
                            "application": "FASTQ Only",
                            "investigatorName": "Investigator",
                            "adapter": "AGATCGGAAGAGCACACGTCTGAACTCCAGTCA",
                            "adapterread2": "AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGT",
                            "workflow": "GenerateFASTQ",
                            "reversecomplement": "0",
                            "iemfileversion": "4",
                            "date": "2015-11-12",
                            "experimentName": "252",
                            "chemistry": "Amplicon",
                            "layoutType": "PAIRED_END"}

        self.assertEqual(correct_metadata, meta_data)


class TestBuildSequencingRunFromSamples(unittest.TestCase):
    """
    Test building the sequencing run from a sample sheet with a csv reader
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_build_valid(self):
        """
        When given a valid directory, ensure a valid SequencingRun is built with Projects, Samples, ect
        :return:
        """
        sheet_file = path.join(path_to_module, "fake_nextseq_run",
                               "SampleSheetClassic.csv")
        meta_data = sample_parser.parse_metadata(sheet_file)

        sequencing_run_type = 'nextseq'

        sequencing_run = sample_parser.build_sequencing_run_from_samples(sheet_file, meta_data, sequencing_run_type)

        # Returns a SequencingRun
        self.assertEqual(type(sequencing_run), model.SequencingRun)
        # Includes two projects
        self.assertEqual(len(sequencing_run.project_list), 2)
        # is of type Project
        self.assertEqual(type(sequencing_run.project_list[0]), model.Project)
        self.assertEqual(type(sequencing_run.project_list[1]), model.Project)
        # Projects have 1 sample each
        self.assertEqual(len(sequencing_run.project_list[0].sample_list), 1)
        self.assertEqual(len(sequencing_run.project_list[1].sample_list), 1)
        # samples are of type Sample
        self.assertEqual(type(sequencing_run.project_list[0].sample_list[0]), model.Sample)
        self.assertEqual(type(sequencing_run.project_list[1].sample_list[0]), model.Sample)
        # samples have SequenceFile
        self.assertEqual(type(sequencing_run.project_list[0].sample_list[0].sequence_file), model.SequenceFile)
        self.assertEqual(type(sequencing_run.project_list[1].sample_list[0].sequence_file), model.SequenceFile)


class TestGetCsvReader(unittest.TestCase):
    """
    Test that the csv reader behaves as expected
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_get_csv_reader_valid_sheet(self):
        """
        Given a valid sample sheet, ensure the parsed sheet matches expected output
        :return:
        """
        sheet_file = path.join(path_to_module, "fake_nextseq_run",
                               "SampleSheetClassic.csv")

        lines = parsers.common.get_csv_reader(sheet_file)

        correct_lines = [
            ['[Header]'],
            ['IEMFileVersion', '5'],
            ['Investigator Name', 'Ikari Yui'],
            ['Experiment Name', 'NGS-Adam'],
            ['Date', '2000-09-13'],
            ['Workflow', 'GenerateFASTQ'],
            ['Application', 'NextSeq FASTQ Only'],
            ['Instrument Type', 'NextSeq/MiniSeq'],
            ['Assay', 'Nextera XT'],
            ['Index Adapters', 'Nextera XT v2 Index Kit A'],
            ['Description', 'Contact Experiment'],
            ['Chemistry', 'Amplicon'],
            [],
            ['[Reads]'],
            ['151'],
            ['151'],
            [],
            ['[Settings]'],
            ['Adapter', 'CTGTCTCTTATACACATCT'],
            [],
            ['[Data]'],
            ['Sample_ID', 'Sample_Name', 'Sample_Plate', 'Sample_Well', 'I7_Index_ID', 'index', 'I5_Index_ID', 'index2',
             'Sample_Project', 'Description'],
            ['SA20121712', 'SA20121712', 'NGS-001 Plate FNC-7', 'A01', 'N701', 'TAAGGCGA', 'S502', 'ATAGAGAG', '67', 'Code Blue'],
            ['SA20121716', 'SA20121716', 'NGS-001 Plate FNC-7', 'A02', 'N702', 'CGTACTAG', 'S502', 'ATAGAGAG', '68', 'Code Blue']
        ]

        for line, c_line in zip(lines, correct_lines):
            self.assertEqual(line, c_line)

    def test_get_csv_reader_no_sheet(self):
        """
        Make sure an error is raised if the csv reader is not given a valid sample sheet
        :return:
        """
        sheet_file = path.join(path_to_module, "fake_nextseq_run",
                               "Data")

        with self.assertRaises(SampleSheetError):
            parsers.common.get_csv_reader(sheet_file)


class TestValidatePfList(unittest.TestCase):
    """
    Testing the various cases that can appear when validating the pf list
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_valid_lists(self):
        self.assertTrue(sample_parser._validate_pf_list(
            ['SA20121712_S2_R1_001.fastq.qz', 'SA20121712_S2_R2_001.fastq.qz']
        ))
        self.assertTrue(sample_parser._validate_pf_list(
            ['SA20121712_S2_R1_001.fastq.qz']
        ))

    def test_invalid_lists(self):
        self.assertFalse(sample_parser._validate_pf_list(
            ['SA20121712_S2_R1_001.fastq.qz', 'SA20121712_S2_R2_001.fastq.qz', 'SA20121712_S2_R3_001.fastq.qz']
        ))

        self.assertFalse(sample_parser._validate_pf_list([]))

    def test_invalid_file_names(self):
        self.assertFalse(sample_parser._validate_pf_list(
            ['SA20121712_S2_R_001.fastq.qz', 'SA20121712_S2_R_001.fastq.qz']
        ))
        self.assertFalse(sample_parser._validate_pf_list(
            ['SA20121712_S2_1_001.fastq.qz', 'SA20121712_S2_2_001.fastq.qz']
        ))


class TestParseSampleList(unittest.TestCase):
    """
    test parsing the list of samples from a sample sheet
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_valid(self):
        """
        Ensure a a parsed valid directory matches the expected sample list
        :return:
        """
        sheet_file = path.join(path_to_module, "fake_nextseq_run",
                               "SampleSheetClassic.csv")

        sample = model.Sample(
            "SA20121712",
            "Code Blue",
            1,
            {
                "Sample_Well": "A01",
                "index": "TAAGGCGA",
                "Sample_Plate": "NGS-001 Plate FNC-7",
                "I7_Index_ID": "N701",
                "sample_project": "67",
                "sequencer_sample_name": "SA20121712",
                "I5_Index_ID": "S502",
                "index2": "ATAGAGAG",
            }
        )

        sequence_file_properties = {
            'Sample_Plate': "NGS-001 Plate FNC-7",
            'Sample_Well': 'A01',
            'I7_Index_ID': 'N701',
            'index': 'TAAGGCGA',
            'I5_Index_ID': 'S502',
            'index2': 'ATAGAGAG'
        }

        file_path_1 = path.join(path_to_module,
                                "fake_nextseq_run", "Data", "Intensities", "BaseCalls", "67",
                                "SA20121712_S2_R1_001.fastq.qz")
        file_path_2 = path.join(path_to_module,
                                "fake_nextseq_run", "Data", "Intensities", "BaseCalls", "67",
                                "SA20121712_S2_R2_001.fastq.qz")
        file_list = [file_path_1, file_path_2]

        res = sample_parser._parse_sample_list(sheet_file)

        # Check sample is the same
        self.assertEqual(res[0].get_uploadable_dict(), sample.get_uploadable_dict())
        # Check sequencing file is correct
        self.assertEqual(res[0].sequence_file.properties_dict, sequence_file_properties)
        self.assertEqual(res[0].sequence_file.file_list.sort(), file_list.sort())

    def test_not_pf_list(self):
        """
        No Valid files were found with names given in sample sheet
        :return:
        """
        directory = path.join(path_to_module, "ngs_not_pf_list")
        file_path = path.join(directory, "SampleSheetClassic.csv")

        with self.assertRaises(SequenceFileError):
            res = sample_parser._parse_sample_list(file_path)

    def test_not_valid_pf_list(self):
        """
        The file list in the sample sheet is invalid
        :return:
        """
        directory = path.join(path_to_module, "ngs_not_valid_pf_list")
        file_path = path.join(directory, "SampleSheetClassic.csv")

        with self.assertRaises(SequenceFileError):
            res = sample_parser._parse_sample_list(file_path)

    def test_space_in_sample_name(self):
        directory = path.join(path_to_module, "ngs_space_in_sample_name")
        file_path = path.join(directory, "SampleSheetClassic.csv")

        # Just making sure this doesn't thow an error
        sample_parser._parse_sample_list(file_path)


class TestParseSamples(unittest.TestCase):
    """
    Test validity or invalidity of parsed samples
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_parse_samples_valid(self):
        """
        Ensure a a parsed valid directory matches the expected samples
        :return:
        """
        sheet_file = path.join(path_to_module, "fake_nextseq_run",
                               "SampleSheetClassic.csv")

        sample1 = model.Sample(
            "SA20121712",
            "Code Blue",
            1,
            {
                "Sample_Well": "A01",
                "index": "TAAGGCGA",
                "Sample_Plate": "NGS-001 Plate FNC-7",
                "I7_Index_ID": "N701",
                "sample_project": "67",
                "sequencer_sample_name": "SA20121712",
                "I5_Index_ID": "S502",
                "index2": "ATAGAGAG",
            }
        )

        sample2 = model.Sample(
            "SA20121716",
            "Code Blue",
            2,
            {
                "Sample_Well": "A02",
                "index": "CGTACTAG",
                "Sample_Plate": "NGS-001 Plate FNC-7",
                "I7_Index_ID": "N702",
                "sample_project": "68",
                "sequencer_sample_name": "SA20121716",
                "I5_Index_ID": "S502",
                "index2": "ATAGAGAG",
            }
        )

        correct_samples = [sample1, sample2]

        res = sample_parser._parse_samples(sheet_file)
        for r_sample, c_sample in zip(res, correct_samples):
            self.assertEqual(r_sample.get_uploadable_dict(), c_sample.get_uploadable_dict())


class TestParseOutSequenceFile(unittest.TestCase):
    """
    test building the sample object
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_parse_out_sequence_file(self):
        """
        Tests that parse out sequence file correctly filters sample related data from the extra params dict
        And ensures that the uploadable dict correctly includes all the needed data after removal
        :return:
        """
        sample = model.Sample(
            "03-3333",
            "Deadly bug",
            None,
            {
                "Sample_Well": "03",
                "index": "CCCCCCCC",
                "Sample_Plate": "3",
                "I7_Index_ID": "N03",
                "sampleName": "03-3333",
                "sampleProject": "6",
                "sequencerSampleId": "03-3333",
                "I5_Index_ID": "S03",
                "index2": "GGGGGGGG",
                "description": "Deadly bug"
            }
        )

        uploadable_dict = {'Sample_Well': '03',
                           'index': 'CCCCCCCC',
                           'Sample_Plate': '3',
                           'I7_Index_ID': 'N03',
                           'sampleName': '03-3333',
                           'sampleProject': '6',
                           'sequencerSampleId': '03-3333',
                           'I5_Index_ID': 'S03',
                           'index2': 'GGGGGGGG',
                           'description': 'Deadly bug'}

        sequence_file_dict = {'Sample_Well': '03',
                              'index': 'CCCCCCCC',
                              'Sample_Plate': '3',
                              'I7_Index_ID': 'N03',
                              'sampleProject': '6',
                              'sequencerSampleId': '03-3333',
                              'I5_Index_ID': 'S03',
                              'index2': 'GGGGGGGG'}

        res = sample_parser._parse_out_sequence_file(sample)

        self.assertEqual(sample.get_uploadable_dict(), uploadable_dict)
        self.assertEqual(res, sequence_file_dict)



"""
This file handles selecting a parser, and parsing/validation with it
"""
import logging

import iridauploader.config as config
import iridauploader.parsers as parsers
from iridauploader.core import model_validator, file_size_validator, uniform_file_count_validator


def get_parser_from_config():
    """
    Uses the 'parser' field in the config file to find and return the parser
    :return:
    """
    parser_instance = config.read_config_option("parser")
    return parsers.parser_factory(parser_instance)


def parse_and_validate(directory):
    """
    Parses and validates a run from a directory
    Throws DirectoryError or ValidationError if invalid

    :param directory:
    :return: a valid sequencing run
    """
    logging.info("*** Beginning Parsing ***")
    parser_instance = get_parser_from_config()

    try:
        sample_sheet = parser_instance.get_sample_sheet(directory)
    except parsers.exceptions.DirectoryError as e:
        logging.debug("parsing_handler:Exception while getting sample sheet from directory")
        raise e

    try:
        sequencing_run = parser_instance.get_sequencing_run(sample_sheet)
    except parsers.exceptions.ValidationError as e:
        logging.debug("parsing_handler:Exception while getting sequencing run with sample_sheet")
        raise e

    logging.info("Validating sequencing run structure")
    validation_result = model_validator.validate_sequencing_run(sequencing_run)
    if not validation_result.is_valid():
        logging.info("parsing_handler:Exception while validating Sequencing Run")
        raise parsers.exceptions.ValidationError("Sequencing Run is not valid", validation_result)

    logging.info("Validating file counts match sequencing run")
    validation_result = uniform_file_count_validator.validate_uniform_file_count(sequencing_run)
    if not validation_result.is_valid():
        logging.info("parsing_handler:Exception while validating Sequencing Run")
        raise parsers.exceptions.ValidationError("Sequencing Run is not valid", validation_result)

    logging.info("Validating files contain data")
    validation_result = file_size_validator.validate_file_size_minimum(sequencing_run)
    if not validation_result.is_valid():
        logging.info("parsing_handler:Exception while validating Sequencing Run")
        raise parsers.exceptions.ValidationError("Sequencing Run is not valid", validation_result)

    logging.info("*** Parsing Done ***")

    return sequencing_run


def get_run_status(directory):
    """
    Given a run directory, returns a DirectoryStatus object created by the parser
    :param directory:
    :return: DirectoryStatus
    """
    parser_instance = get_parser_from_config()
    status = parser_instance.find_single_run(directory)
    return status


def get_run_status_list(batch_directory):
    """
    Given a directory containing potential runs, returns a list ofDirectoryStatus objects created by the parser
    :param batch_directory:
    :return: list of DirectoryStatus objects
    """
    parser_instance = get_parser_from_config()
    status_list = parser_instance.find_runs(batch_directory)
    return status_list



import sys
# PyQt needs to be imported like this because for whatever reason they decided not to include a __all__ = [...]
import PyQt5.QtWidgets as QtWidgets

from appdirs import user_log_dir
import os
import traceback

import iridauploader.gui.main_dialog as main_dialog


def get_crash_log_file_path():
    # Normal base logging directory name
    log_directory_name = "irida-uploader"
    # Use systems default logging path, and append our named directory
    log_file_path = os.path.join(user_log_dir(log_directory_name), 'crash.log')
    # Create the directory if it doesn't exist
    if not os.path.exists(user_log_dir(log_directory_name)):
        os.makedirs(user_log_dir(log_directory_name))

    return log_file_path


def except_hook(error_exctype, error_value, error_traceback):
    # This function is used to override the default __excepthook__ so that we can write our crash log to file

    # keep existing functionality of __excepthook__
    sys.__excepthook__(error_exctype, error_value, error_traceback)

    # get a formatted string with out traceback and print it to our crash log file
    tb = ''.join(traceback.format_tb(error_traceback))
    f = open(get_crash_log_file_path(), "a+")
    f.write(tb)

    # after an exception like this, exit with error
    sys.exit(1)


def main():
    """
    Entry point for GUI
    :return:
    """
    app = QtWidgets.QApplication(["IRIDA Uploader"])
    dlg = main_dialog.MainDialog()
    dlg.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    # override except hook for custom crash log
    sys.excepthook = except_hook
    main()



class ValidationError(Exception):
    """
    An exception to be raised validation errors happen while parsing

    includes a ValidationResult object with list of errors that occurred
    """
    def __init__(self, message, validation_result):
        """Initialize a SampleError.

        Args:
            message: the summary message that's causing the error.
            validation_result: a ValidationResult object with list of errors that occurred
        """
        self._message = message
        self._validation_result = validation_result

    @property
    def message(self):
        return self._message

    @property
    def validation_result(self):
        return self._validation_result

    def __str__(self):
        return self.message



import logging
# PyQt needs to be imported like this because for whatever reason they decided not to include a __all__ = [...]
import PyQt5.QtCore as QtCore
import PyQt5.QtWidgets as QtWidgets

from iridauploader.config import config
from iridauploader.parsers import supported_parsers

from iridauploader.gui import tools, colours


class ConfigDialog(QtWidgets.QDialog):
    """
    Configuration/Settings dialog window.

    This has multiple options that the user can change instead of using the config file.
    On opening the dialog the settings will be loaded from the config file.
    When the user accepts, or clicks "save and test", the settings will be written to the config file.
    """
    def __init__(self, parent=None):
        super().__init__(parent)
        logging.debug("GUI: Setting up ConfigDialog")

        # Set dialog title
        self.setWindowTitle("Configure Settings")

        # Widgets
        self._btn_import_settings = QtWidgets.QPushButton("Import Configuration From File")
        self._client_id_label = QtWidgets.QLabel("Client ID")
        self._client_id = QtWidgets.QLineEdit()
        self._client_secret_label = QtWidgets.QLabel("Client Secret")
        self._client_secret = QtWidgets.QLineEdit()
        self._username_label = QtWidgets.QLabel("Username")
        self._username = QtWidgets.QLineEdit()
        self._password_label = QtWidgets.QLabel("Password")
        self._password = QtWidgets.QLineEdit()
        self._password.setEchoMode(QtWidgets.QLineEdit.Password)
        self._base_url_label = QtWidgets.QLabel("Base URL")
        self._base_url = QtWidgets.QLineEdit()
        self._parser_label = QtWidgets.QLabel("Parser")
        self._parser = QtWidgets.QComboBox()
        self._parser.addItems(supported_parsers)
        self._read_only_mode = QtWidgets.QCheckBox()
        self._read_only_mode_label = QtWidgets.QLabel("Read Only Mode")
        self._timeout_label = QtWidgets.QLabel("(ADVANCED) Timeout Seconds Per MB")
        self._timeout = QtWidgets.QSpinBox()
        self._min_file_size_label = QtWidgets.QLabel("(ADVANCED) Minimum File Size per file (KB)")
        self._min_file_size = QtWidgets.QSpinBox()
        self._min_file_size.setRange(0, 2147483647)  # set max as max allowed by QSpinBox

        self._btn_check_settings = QtWidgets.QPushButton("Save and Test Settings")
        self._settings_status = QtWidgets.QPlainTextEdit()
        self._settings_status.setReadOnly(True)
        self._settings_status.setEnabled(False)
        self._settings_status.setStyleSheet("color: black")
        self._btn_accept = QtWidgets.QPushButton("Accept")
        self._btn_cancel = QtWidgets.QPushButton("Cancel")

        # connections
        self._btn_import_settings.clicked.connect(self._import_config_from_file)
        self._btn_check_settings.clicked.connect(self._check_connection_to_irida)
        self._btn_accept.clicked.connect(self._accept_and_exit)
        self._btn_cancel.clicked.connect(self._cancel_and_exit)

        # Set Layout and Geometry
        self.setLayout(self._layout())
        self.setGeometry(0, 0, 600, 300)

        # Init settings
        self._get_settings_from_file()
        self._contact_irida()

        # Error message popup
        self._error_message_box = QtWidgets.QMessageBox(parent=self)
        self._error_message_box.setText(
            'Config file was invalid. Please check your file and try again.<br>'
            '<a href="https://irida-uploader.readthedocs.io/en/stable/configuration.html">'
            'Click here for details on config files</a>'
        )
        self._error_message_box.setIcon(QtWidgets.QMessageBox.Critical)

    def _layout(self):
        """
        Layout widgets
        :return: QtWidgets.QVBoxLayout()
        """
        # base layout
        layout = QtWidgets.QVBoxLayout()
        # Load from file
        layout.addWidget(self._btn_import_settings)
        # Client ID
        client_id_layout = QtWidgets.QHBoxLayout()
        client_id_layout.addWidget(self._client_id_label)
        client_id_layout.addWidget(self._client_id)
        layout.addLayout(client_id_layout)
        # Client Secret
        client_secret_layout = QtWidgets.QHBoxLayout()
        client_secret_layout.addWidget(self._client_secret_label)
        client_secret_layout.addWidget(self._client_secret)
        layout.addLayout(client_secret_layout)
        # Username
        username_layout = QtWidgets.QHBoxLayout()
        username_layout.addWidget(self._username_label)
        username_layout.addWidget(self._username)
        layout.addLayout(username_layout)
        # Password
        password_layout = QtWidgets.QHBoxLayout()
        password_layout.addWidget(self._password_label)
        password_layout.addWidget(self._password)
        layout.addLayout(password_layout)
        # Base URL
        base_url_layout = QtWidgets.QHBoxLayout()
        base_url_layout.addWidget(self._base_url_label)
        base_url_layout.addWidget(self._base_url)
        layout.addLayout(base_url_layout)
        # Parser
        parser_layout = QtWidgets.QHBoxLayout()
        parser_layout.addWidget(self._parser_label)
        parser_layout.addWidget(self._parser)
        layout.addLayout(parser_layout)
        # Read Only
        read_only_layout = QtWidgets.QHBoxLayout()
        read_only_layout.addWidget(self._read_only_mode_label)
        read_only_layout.addWidget(self._read_only_mode)
        layout.addLayout(read_only_layout)
        # Timeout
        timeout_layout = QtWidgets.QHBoxLayout()
        timeout_layout.addWidget(self._timeout_label)
        timeout_layout.addWidget(self._timeout)
        layout.addLayout(timeout_layout)
        # Min file size
        min_file_size_layout = QtWidgets.QHBoxLayout()
        min_file_size_layout.addWidget(self._min_file_size_label)
        min_file_size_layout.addWidget(self._min_file_size)
        layout.addLayout(min_file_size_layout)
        # Buttons and status
        layout.addWidget(self._btn_check_settings)
        layout.addWidget(self._settings_status)
        button_layout = QtWidgets.QHBoxLayout()
        button_layout.addWidget(self._btn_accept)
        button_layout.addWidget(self._btn_cancel)
        layout.addLayout(button_layout)

        return layout

    def _accept_and_exit(self):
        """
        Writes the settings to file and then closes
        :return:
        """
        self._write_settings_to_file()
        self._contact_irida()
        self.close()

    def _cancel_and_exit(self):
        """
        closes this widget
        :return:
        """
        self.close()

    def _get_settings_from_file(self):
        """
        Loads the config settings from file and fills the widgets
        :return:
        """
        config.setup()
        self._client_id.setText(config.read_config_option('client_id'))
        self._client_secret.setText(config.read_config_option('client_secret'))
        self._username.setText(config.read_config_option('username'))
        self._password.setText(config.read_config_option('password'))
        self._base_url.setText(config.read_config_option('base_url'))
        index = self._parser.findText(config.read_config_option('parser'))
        self._parser.setCurrentIndex(index)
        read_only_bool = config.read_config_option("readonly", bool, False)
        read_only_state = QtCore.Qt.Checked if read_only_bool else QtCore.Qt.Unchecked
        self._read_only_mode.setCheckState(read_only_state)
        timeout_int = config.read_config_option("timeout", int, 10)
        self._timeout.setValue(timeout_int)
        min_file_size_int = config.read_config_option("minimum_file_size", int, 0)
        self._min_file_size.setValue(min_file_size_int)

    def _write_settings_to_file(self):
        """
        Writes the settings in the config boxes to file
        :return:
        """
        config.set_config_options(client_id=self._client_id.text(),
                                  client_secret=self._client_secret.text(),
                                  username=self._username.text(),
                                  password=self._password.text(),
                                  base_url=self._base_url.text(),
                                  parser=self._parser.currentText(),
                                  readonly=self._read_only_mode.isChecked(),
                                  timeout=self._timeout.value(),
                                  minimum_file_size=self._min_file_size.value())
        config.write_config_options_to_file()

    def _contact_irida(self):
        """
        Attempts to connect to IRIDA
        Sets the style and text of the status widget to green/red to indicate connected/error
        :return:
        """
        try:
            tools.is_connected_to_irida()
            self._settings_status.setPlainText("Connection OK")
            self._settings_status.setStyleSheet("background-color: {}; color: black;".format(colours.GREEN_LIGHT))
            logging.info("Successfully connected to IRIDA")
        except Exception as e:
            self._settings_status.setPlainText("ERROR: " + str(e))
            self._settings_status.setStyleSheet("background-color: {}; color: black;".format(colours.RED_LIGHT))
            logging.info("Error occurred while trying to connect to IRIDA")

    def _check_connection_to_irida(self):
        """
        Writes the settings in the boxes to the config file, and then tries to connect to IRIDA
        :return:
        """
        self._write_settings_to_file()
        self._contact_irida()

    def center_window(self):
        """
        sets the dialog position to the center of the parent
        :return:
        """
        qt_rectangle = self.frameGeometry()
        parent_geo = self.parent().frameGeometry().center()
        qt_rectangle.moveCenter(parent_geo)
        self.move(qt_rectangle.topLeft())

    def _import_config_from_file(self):
        """
        Opens dialog to choose file, exits early if no file is chosen or file is invalid
        Then, the files parameters are loaded and written to the default config file
        :return:
        """
        logging.debug("GUI CONFIG: _import_config_from_file clicked")
        # Open a file open dialog and get returned string
        config_file_path = QtWidgets.QFileDialog.getOpenFileName(self,
                                                                 caption="Select Uploader Config File",
                                                                 filter="*.conf")[0]
        logging.debug("GUI CONFIG: result: " + config_file_path)
        # If user exits window without selecting a file, return early
        if config_file_path == "":
            logging.debug("GUI CONFIG: user canceled opening file")
            return

        # Do a first pass on the loaded file to see if it is valid.
        # If there is no exception, the file can be loaded normally.
        try:
            config.read_config_file_into_parser(config_file_path)
        except Exception:
            self._error_message_box.exec()
            return

        # set the config file to be the file the user selected
        config.set_config_file(config_file_path)
        # Load the properties into the window
        self._get_settings_from_file()
        # Reset the config file back to use default
        config.set_config_file(None)
        config.setup()
        # Write the settings from the user selected file to the default file
        self._write_settings_to_file()
        # Check connection with IRIDA
        self._contact_irida()
